# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
from typing import Optional, List, Tuple, Any
from itertools import repeat

from torch.utils.data import Dataset as TorchDataset
from tqdm import tqdm
import numpy as np

from .timeseries import TimeSeries, TimeSeriesCorpus


class WindowsDataset(TorchDataset):
    def __init__(
        self,
        corpus: TimeSeriesCorpus,
        windows: List[Tuple[int, int, int]],
    ) -> None:
        super(WindowsDataset, self).__init__()
        self.corpus = corpus
        self.windows = windows

    def __len__(self) -> int:
        return len(self.windows)

    def _fetch(self, index: int) -> Tuple[TimeSeries, slice]:
        scope_id, start, length = self.windows[index]
        window = slice(start, start + length)
        return self.corpus[scope_id], window

    def __getitem__(self, index: int) -> TimeSeries:
        ts, window = self._fetch(index)
        subseq = ts[window]
        return subseq

    @classmethod
    def sliding_windows(
        cls,
        corpus: TimeSeriesCorpus,
        window_size: int,
        window_size_plus: int = 0,
        window_size_minus: int = 0,
        window_shift: int = 1,
        seed: int = 42,
    ) -> WindowsDataset:
        max_window_size = window_size + window_size_plus
        min_window_size = window_size - window_size_minus

        windows = []
        np.random.seed(seed)
        for scope_id, series in tqdm(
            enumerate(corpus),
            desc="Building dataset",
            total=len(corpus),
        ):
            stop = len(series) - min_window_size
            index = np.arange(stop, 0, -window_shift)
            window_length = np.random.randint(
                min_window_size, max_window_size + 1, size=index.shape
            )
            window_length = np.minimum(len(series) - index, window_length)
            windows.extend(
                zip(repeat(scope_id, len(index)), index, window_length)
            )
        return cls(corpus, windows)

    @classmethod
    def random_windows(
        cls,
        corpus: TimeSeriesCorpus,
        n_windows: int,
        window_size: int,
        window_size_plus: int = 0,
        window_size_minus: int = 0,
        window_shift: int = 1,
        seed: int = 42,
    ) -> WindowsDataset:
        max_window_size = window_size + window_size_plus
        min_window_size = window_size - window_size_minus

        windows = []
        np.random.seed(seed)
        for scope_id, series in tqdm(
            enumerate(corpus),
            desc="Building dataset",
            total=len(corpus),
        ):
            stop = len(series) - min_window_size
            index = np.arange(stop, 0, -window_shift)
            window_length = np.random.randint(
                min_window_size, max_window_size + 1, size=index.shape
            )
            window_length = np.minimum(len(series) - index, window_length)
            windows.extend(
                zip(repeat(scope_id, len(index)), index, window_length)
            )
        window_idx = np.random.choice(
            len(windows),
            size=n_windows,
            replace=len(windows) < n_windows,
        )
        windows = [windows[i] for i in window_idx]
        return cls(corpus, windows)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from src.gluonts.dataset.repository.datasets import get_dataset
from src.gluonts.mx.trainer import Trainer

from src.gluonts.nursery.temporal_hierarchical_forecasting.model.cop_deepar import (
    COPDeepAREstimator,
)
from src.gluonts.nursery.temporal_hierarchical_forecasting.eval.evaluation import (
    evaluate_predictor,
)


EVALUATE_ALL_LEVELS = True


def main():
    dataset = get_dataset("exchange_rate", regenerate=False)

    estimator = COPDeepAREstimator(
        freq=dataset.metadata.freq,
        prediction_length=dataset.metadata.prediction_length,
        base_estimator_name="DeepAREstimatorForCOP",
        base_estimator_hps={},
        trainer=Trainer(
            epochs=100,
            hybridize=False,
        ),
    )

    predictor = estimator.train(dataset.train)

    results = evaluate_predictor(
        predictor=predictor,
        test_dataset=dataset.test,
        evaluate_all_levels=EVALUATE_ALL_LEVELS,
        freq=dataset.metadata.freq,
    )

    print(results)


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations

from typing import (
    List,
    Tuple,
    Dict,
    Union,
    Sequence,
    Optional,
    Callable,
    overload,
)
from operator import eq
from collections import defaultdict
import warnings

from sklearn.preprocessing import LabelEncoder, StandardScaler
from numpy import ndarray
from tqdm import tqdm
import numpy as np
import pandas as pd
import torch as pt


def _dict_equal(dict1: Dict, dict2: Dict, cmp_func: Callable = eq) -> bool:
    if dict1.keys() == dict2.keys():
        return all(cmp_func(dict1[k], dict2[k]) for k in dict1.keys())
    return False


def _pad_array_along_axis(
    array: ndarray, pad: Tuple[int, int], axis: int = 0, constant_values=np.nan
) -> ndarray:
    full_pad = [(0, 0)] * array.ndim
    full_pad[axis] = pad
    return np.pad(
        array, full_pad, mode="constant", constant_values=constant_values
    )


class TimeSeriesInstant(object):
    def __init__(
        self,
        target: ndarray,
        timestamp: pd.Timestamp,
        series_name: Optional[str],
        target_names: List[str],
    ):
        if target.ndim == 1:
            self.target = target
        else:
            raise ValueError("target has to be a 1-d array")
        self.timestamp = timestamp
        self.series_name = series_name
        self.target_names = target_names

        self._features = dict()
        self._categorical_features = set()
        self._numerical_features = set()

    def add_features(self, name: str, data: ndarray, data_type: str) -> None:
        valid_data_type = ["cat", "num", "categorical", "numerical"]
        data_type = data_type.lower()
        if data_type not in valid_data_type:
            raise ValueError(f"data_type should be one of {valid_data_type}")
        numerical = data_type in ["num", "numerical"]

        valid_ndim = 1 if numerical else 0
        if data.ndim != valid_ndim:
            raise ValueError(
                f"{feature_type} and {data_type} type feature requires {valid_ndim}-d data"
            )

        self._features[name] = data
        if numerical:
            self._numerical_features.add(name)
        else:
            self._categorical_features.add(name)

    @property
    def categorical_features(self) -> Dict:
        return {
            name: self._features[name] for name in self._categorical_features
        }

    @property
    def numerical_features(self) -> Dict:
        return {
            name: self._features[name] for name in self._numerical_features
        }

    @property
    def d_data(self) -> int:
        return self.target.shape[1]

    def __eq__(self, other) -> bool:
        return all(
            [
                isinstance(other, TimeSeriesInstant),
                np.array_equal(self.target, other.target),
                np.array_equal(self.timestamp, other.timestamp),
                self.series_name == other.series_name,
                np.array_equal(self.target_names, other.target_names),
                _dict_equal(
                    self.categorical_features,
                    other.categorical_features,
                    np.array_equal,
                ),
                _dict_equal(
                    self.numerical_features,
                    other.numerical_features,
                    np.array_equal,
                ),
            ]
        )

    def __repr__(self):
        string = f"time = {self.timestamp:%Y-%m-%d %H:%M:%S}\n"
        string += f"data = {self.target}\n"
        for name, val in self.categorical_features.items():
            string += f"{name} = {val}\n"
        for name, val in self.numerical_features.items():
            string += f"{name} = {val}\n"
        return string


class TimeSeries(Sequence[TimeSeriesInstant]):
    def __init__(
        self,
        target: ndarray,
        start_time: Optional[pd.Timestamp] = None,
        time_unit: Optional[str] = None,
        time_index: Optional[pd.DatetimeIndex] = None,
        series_name: Optional[str] = None,
        target_names: Optional[Union[str, List[str]]] = None,
    ):
        if target.ndim == 1:
            self.target = target.reshape(-1, 1)
        else:
            self.target = target
        if time_index is None:
            assert not (
                (start_time is None) or (time_unit is None)
            ), "if `time_index` is not given, both `start_time` and `time_unit` must be specified"
            self.time_index = pd.date_range(
                start=start_time, periods=len(self), freq=time_unit
            )
            self.time_unit = time_unit
        else:
            assert len(time_index) == len(self.target)
            self.time_index = time_index
            self.time_unit = time_unit or pd.infer_freq(time_index)
        self.series_name = series_name or "Unknown"
        if target_names is None:
            self.target_names = [f"d_{i+1}" for i in range(self.d_data)]
        elif isinstance(target_names, str):
            if self.d_data > 1:
                self.target_names = [
                    f"{target_names}_{i+1}" for i in range(self.d_data)
                ]
            else:
                self.target_names = [target_names]
        else:
            err_msg = f"expected {self.d_data}, given {len(target_names)}."
            if len(target_names) > self.d_data:
                raise ValueError(f"Too many target names are given, {err_msg}")
            if len(target_names) < self.d_data:
                raise ValueError(
                    f"Not enough target names are given, {err_msg}"
                )
            self.target_names = target_names

        self._features = dict()
        self._static_features = set()
        self._revealed_features = set()
        self._observed_features = set()
        self._categorical_features = set()
        self._numerical_features = set()

    def add_features(
        self, name: str, data: ndarray, data_type: str, feature_type: str
    ) -> None:
        valid_data_type = ["cat", "num", "categorical", "numerical"]
        valid_feature_type = ["static", "revealed", "observed"]
        data_type = data_type.lower()
        feature_type = feature_type.lower()
        if data_type not in valid_data_type:
            raise ValueError(f"data_type should be one of {valid_data_type}")
        if feature_type not in valid_feature_type:
            raise ValueError(
                f"feature_type should be one of {valid_feature_type}"
            )

        numerical = data_type in ["num", "numerical"]
        static = feature_type == "static"
        revealed = feature_type == "revealed"

        valid_ndim = 1 if numerical else 0
        if not static:
            valid_ndim += 1
        if data.ndim != valid_ndim:
            raise ValueError(
                f"{feature_type} and {data_type} type feature requires {valid_ndim}-d data"
            )

        self._features[name] = data
        if static:
            self._static_features.add(name)
        elif revealed:
            self._revealed_features.add(name)
        else:
            self._observed_features.add(name)

        if numerical:
            self._numerical_features.add(name)
        else:
            self._categorical_features.add(name)

    @property
    def static_categorical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._static_features & self._categorical_features
        }

    @property
    def static_numerical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._static_features & self._numerical_features
        }

    @property
    def revealed_categorical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._revealed_features & self._categorical_features
        }

    @property
    def revealed_numerical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._revealed_features & self._numerical_features
        }

    @property
    def observed_categorical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._observed_features & self._categorical_features
        }

    @property
    def observed_numerical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._observed_features & self._numerical_features
        }

    @property
    def _dynamic_features(self) -> set:
        return self._observed_features | self._revealed_features

    @property
    def dynamic_categorical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._dynamic_features & self._categorical_features
        }

    @property
    def dynamic_numerical_features(self) -> Dict:
        return {
            name: self._features[name]
            for name in self._dynamic_features & self._numerical_features
        }

    def __eq__(self, other):
        return all(
            [
                isinstance(other, TimeSeries),
                np.array_equal(self.target, other.target),
                np.array_equal(self.time_index, other.time_index),
                self.series_name == other.series_name,
                np.array_equal(self.target_names, other.target_names),
                _dict_equal(
                    self.static_categorical_features,
                    other.static_categorical_features,
                    np.array_equal,
                ),
                _dict_equal(
                    self.static_numerical_features,
                    other.static_numerical_features,
                    np.array_equal,
                ),
                _dict_equal(
                    self.revealed_categorical_features,
                    other.revealed_categorical_features,
                    np.array_equal,
                ),
                _dict_equal(
                    self.revealed_numerical_features,
                    other.revealed_numerical_features,
                    np.array_equal,
                ),
                _dict_equal(
                    self.observed_categorical_features,
                    other.observed_categorical_features,
                    np.array_equal,
                ),
                _dict_equal(
                    self.observed_numerical_features,
                    other.observed_numerical_features,
                    np.array_equal,
                ),
            ]
        )

    def __len__(self):
        return len(self.target)

    @overload
    def index_by_timestamp(self, index: pd.Timestamp) -> int: ...

    @overload
    def index_by_timestamp(self, index: List[pd.Timestamp]) -> List[int]: ...

    def index_by_timestamp(self, index):
        return pd.Series(
            index=self.time_index, data=np.arange(len(self.time_index))
        ).loc[index]

    @overload
    def __getitem__(self, index: int) -> TimeSeriesInstant: ...

    @overload
    def __getitem__(self, index: pd.Timestamp) -> TimeSeriesInstant: ...

    @overload
    def __getitem__(self, index: slice) -> TimeSeries: ...

    @overload
    def __getitem__(self, index: List[int]) -> TimeSeries: ...

    @overload
    def __getitem__(self, index: List[pd.Timestamp]) -> TimeSeries: ...

    def __getitem__(self, index):
        if isinstance(index, pd.Timestamp) or (
            isinstance(index, List) and isinstance(index[0], pd.Timestamp)
        ):
            index = self.index_by_timestamp(index)

        left_pad = right_pad = None
        if isinstance(index, slice):
            if (index.stop is not None) and (index.stop > 0):
                stop = index.stop
                if stop > len(self):
                    right_pad = stop - len(self)
                    stop = None
                start = index.start
                if (index.start is not None) and (index.start < 0):
                    left_pad = -index.start
                    start = None
                index = slice(start, stop)
        target = self.target[index]
        time_index = self.time_index[index]

        if isinstance(index, int):
            obj = TimeSeriesInstant(
                target,
                time_index,
                series_name=self.series_name,
                target_names=self.target_names,
            )
            for name in self._categorical_features:
                if name in self._static_features:
                    data = self._features[name]
                else:
                    data = self._features[name][index]
                obj.add_features(name, data, data_type="cat")
            for name in self._numerical_features:
                if name in self._static_features:
                    data = self._features[name]
                else:
                    data = self._features[name][index]
                obj.add_features(name, data, data_type="num")
            return obj
        else:
            if left_pad is not None:
                if self.time_unit is not None:
                    td = pd.Timedelta(1, unit=self.time_unit).to_numpy()
                else:
                    if len(time_index) < 2:
                        raise IndexError(
                            "Cannot pad time index with a single timestamp"
                        )
                    td = time_index[1] - time_index[0]
                time_index = np.r_[
                    np.arange(
                        time_index[0] - td * left_pad, time_index[0], td
                    ),
                    time_index,
                ]
                target = _pad_array_along_axis(
                    target, (left_pad, 0), constant_values=np.nan
                )
            if right_pad is not None:
                if self.time_unit is not None:
                    td = pd.Timedelta(1, unit=self.time_unit).to_numpy()
                else:
                    if len(time_index) < 2:
                        raise IndexError(
                            "Cannot pad time index with a single timestamp"
                        )
                    td = time_index[-1] - time_index[-2]
                time_index = np.r_[
                    time_index,
                    np.arange(
                        time_index[-1] + td,
                        time_index[-1] + td * (right_pad + 1),
                        td,
                    ),
                ]
                target = _pad_array_along_axis(
                    target, (0, right_pad), constant_values=np.nan
                )
            obj = TimeSeries(
                target,
                time_index=time_index,
                series_name=self.series_name,
                target_names=self.target_names,
            )
            for name in self._static_features:
                data_type = (
                    "cat" if name in self._categorical_features else "num"
                )
                data = self._features[name]
                obj.add_features(
                    name, data, data_type=data_type, feature_type="static"
                )
            for name in self._revealed_features | self._observed_features:
                data_type = (
                    "cat" if name in self._categorical_features else "num"
                )
                feature_type = (
                    "revealed"
                    if name in self._revealed_features
                    else "observed"
                )
                val = -1 if data_type == "cat" else np.nan
                data = self._features[name][index]
                if left_pad is not None:
                    data = _pad_array_along_axis(
                        data, (left_pad, 0), constant_values=val
                    )
                if right_pad is not None:
                    data = _pad_array_along_axis(
                        data, (0, right_pad), constant_values=val
                    )
                obj.add_features(
                    name, data, data_type=data_type, feature_type=feature_type
                )
            return obj

    @property
    def d_data(self) -> int:
        return self.target.shape[1]

    @classmethod
    def from_pandas_series(
        cls, series: pd.Series, target_name: Optional[str] = None
    ):
        return cls(
            series.values,
            time_index=series.index,
            series_name=series.name,
            target_names=target_name,
        )

    def to_pandas_dataframe(self) -> pd.DataFrame:
        df = pd.DataFrame(
            self.target, columns=self.target_names, index=self.time_index
        )
        for name in self._static_features:
            data = self._features[name]
            if data.ndim == 0 or (data.ndim == 1 and data.shape[0] == 1):
                df[name] = np.tile(data, len(self))
            else:
                for i in range(data.shape[0]):
                    df[f"{name}_{i}"] = np.tile(data[i], len(self))
        for name in self._revealed_features | self._observed_features:
            data = self._features[name]
            if data.ndim == 1:
                df[name] = data
            elif data.ndim == 2 and data.shape[1] == 1:
                df[name] = data[:, 0]
            else:
                for i in range(data.shape[0]):
                    df[f"{name}_{i}"] = data[:, i]
        return df

    def to_pandas_series(self) -> pd.Series:
        df = self.to_pandas_dataframe()
        series = df[df.columns[0]]
        series.name = "1D-data"
        return series

    def __repr__(self) -> str:
        string = f"Time series {self.series_name} of size {len(self)}x{self.d_data}\n"
        string += f"static_features: {self._static_features}\n"
        string += f"revealed_features: {self._revealed_features}\n"
        string += f"observed_features: {self._observed_features}\n"
        string += f"categorical_features: {self._categorical_features}\n"
        string += f"numerical_features: {self._numerical_features}\n"
        return string


class TimeSeriesCorpus(Sequence[TimeSeries]):
    def __init__(
        self,
        instances: List[TimeSeries],
        scopes: Optional[List[Tuple[int, slice]]] = None,
        categorical_encoders: Optional[defaultdict] = None,
        numerical_encoders: Optional[defaultdict] = None,
        add_series_id: bool = False,
    ) -> None:
        if categorical_encoders is None and numerical_encoders is None:
            self.categorical_encoders = defaultdict(LabelEncoder)
            self.numerical_scalers = defaultdict(StandardScaler)
            self.instances = self._check_consistency(instances)
        else:
            self.categorical_encoders = categorical_encoders
            self.numerical_scalers = numerical_encoders
            self.instances = instances

        if scopes is None:
            self.scopes = [
                (ts_id, slice(len(ts)))
                for ts_id, ts in enumerate(self.instances)
            ]
        else:
            self.scopes = scopes
        self.add_series_id = add_series_id

    def _check_consistency(
        self, instances: List[TimeSeries]
    ) -> List[TimeSeries]:
        def _consistent(ts1: TimeSeries, ts2: TimeSeries) -> bool:
            return all(
                [
                    np.array_equal(ts1.target_names, ts2.target_names),
                    ts1._static_features == ts2._static_features,
                    ts1._revealed_features == ts2._revealed_features,
                    ts1._observed_features == ts2._observed_features,
                    ts1._categorical_features == ts2._categorical_features,
                    ts1._numerical_features == ts2._numerical_features,
                ]
            )

        cats = defaultdict(list)
        nums = defaultdict(list)
        ref = instances[0]
        for i, ts in enumerate(instances):
            if not _consistent(ref, ts):
                raise ValueError("Given time series are not consistent")
            for name in ts._categorical_features:
                cats[name].append(ts._features[name])
            for name in ts._numerical_features & ts._static_features:
                nums[name].append(ts._features[name])
        for name, vals in cats.items():
            data = np.stack(vals, axis=0).flatten()
            self.categorical_encoders[name].fit(data)
        for name, vals in nums.items():
            data = np.stack(vals, axis=0)
            self.numerical_scalers[name].fit(data)
        return instances

    def __len__(self) -> int:
        return len(self.scopes)

    def __getitem__(self, index: int) -> TimeSeries:
        ts_id, scope = self.scopes[index]
        ts = self.instances[ts_id][scope]
        for name in ts._categorical_features:
            data = ts._features[name]
            shape = data.shape
            data = self.categorical_encoders[name].transform(data.flatten())
            ts._features[name] = data.reshape(shape)
        for name in ts._numerical_features & ts._static_features:
            data = ts._features[name]
            shape = data.shape
            if data.ndim == 1:
                data = data[None, :]
            data = self.numerical_scalers[name].transform(data)
            ts._features[name] = data.reshape(shape)
        if self.add_series_id:
            ts.add_features(
                "id", np.array(ts_id), data_type="cat", feature_type="static"
            )
        return ts

    def __getattr__(self, key: str):
        value = getattr(
            super(TimeSeriesCorpus, self).__getattribute__("instances")[0], key
        )
        if key in ["_categorical_features", "_static_features"]:
            if self.add_series_id:
                value = value | {"id"}
        return value

    @property
    def cardinalities(self) -> Dict:
        cardi = {
            name: len(enc.classes_)
            for name, enc in self.categorical_encoders.items()
        }
        if self.add_series_id:
            cardi["id"] = len(self)
        return cardi

    def split_by_timestamp(
        self,
        split_timestamp: pd.Timestamp,
        horizon: Optional[int] = None,
        context: Optional[int] = None,
        n_rolls: int = 1,
        shift: Optional[int] = None,
        raise_error: bool = True,
    ) -> Tuple[TimeSeriesCorpus, TimeSeriesCorpus]:
        shift = shift or horizon
        train_scopes = []
        test_scopes = []
        for ts_id, scope in self.scopes:
            ts = self.instances[ts_id]
            start = scope.start or 0
            stop = scope.stop or len(ts)
            split_index = ts.index_by_timestamp(split_timestamp)
            test_start = start if context is None else split_index - context
            if horizon is None:
                test_stop = stop
            else:
                test_stop = split_index + shift * (n_rolls - 1) + horizon
            if not ((start <= test_start) and (test_stop <= stop)):
                err_msg = (
                    f"Scope {ts_id}, requested {test_start}:{test_stop} is not "
                    f"contained by the current scope {start}:{stop}."
                )
                if raise_error:
                    raise ValueError(err_msg)
                warnings.warn(f"{err_msg} No scope added to test split.")
                train_scopes.append((ts_id, slice(start, stop)))
            else:
                test_scopes.append((ts_id, slice(test_start, test_stop)))
                train_scopes.append((ts_id, slice(start, split_index)))
        train_corpus = TimeSeriesCorpus(
            self.instances,
            train_scopes,
            categorical_encoders=self.categorical_encoders,
            numerical_encoders=self.numerical_scalers,
            add_series_id=self.add_series_id,
        )
        test_corpus = TimeSeriesCorpus(
            self.instances,
            test_scopes,
            categorical_encoders=self.categorical_encoders,
            numerical_encoders=self.numerical_scalers,
            add_series_id=self.add_series_id,
        )
        return train_corpus, test_corpus

    def split_from_end(
        self,
        horizon: int,
        context: Optional[int] = None,
        n_rolls: int = 1,
        shift: Optional[int] = None,
        raise_error: bool = True,
    ) -> Tuple[TimeSeriesCorpus, TimeSeriesCorpus]:
        shift = shift or horizon
        train_scopes = []
        test_scopes = []
        for ts_id, scope in self.scopes:
            new_start = scope.stop - shift * (n_rolls - 1) - horizon
            start = scope.start or 0
            test_start = (
                scope.start if context is None else new_start - context
            )
            if test_start < start:
                err_msg = (
                    f"Scope {ts_id}, requested {test_start}:{scope.stop} is not "
                    f"contained by the current scope {start}:{scope.stop}."
                )
                if raise_error:
                    raise ValueError(err_msg)
                warnings.warn(f"{err_msg} No scope added to test split.")
                train_scopes.append((ts_id, slice(start, scope.stop)))
            else:
                test_scopes.append((ts_id, slice(test_start, scope.stop)))
                train_scopes.append((ts_id, slice(scope.start, new_start)))
        train_corpus = TimeSeriesCorpus(
            self.instances,
            train_scopes,
            categorical_encoders=self.categorical_encoders,
            numerical_encoders=self.numerical_scalers,
            add_series_id=self.add_series_id,
        )
        test_corpus = TimeSeriesCorpus(
            self.instances,
            test_scopes,
            categorical_encoders=self.categorical_encoders,
            numerical_encoders=self.numerical_scalers,
            add_series_id=self.add_series_id,
        )
        return train_corpus, test_corpus

    def split_from_end_by_blocks(
        self,
        horizon: int,
        context: Optional[int] = None,
        n_rolls: int = 1,
        n_blocks: int = 1,
        shift: Optional[int] = None,
        raise_error: bool = True,
    ) -> Tuple[TimeSeriesCorpus, TimeSeriesCorpus]:
        shift = shift or horizon
        train_scopes = []
        test_scopes = []
        # assert self.instances == self.scopes
        for ts_id, scope in self.scopes:
            block_size = len(self.instances[ts_id][scope]) // n_blocks
            ini = scope.start or 0
            for i in range(n_blocks):
                start = ini + i * block_size
                stop = start + block_size
                new_start = stop - shift * (n_rolls - 1) - horizon
                test_start = start if context is None else new_start - context
                if test_start < start:
                    err_msg = (
                        f"Scope {ts_id}, requested {test_start}:{stop} is not "
                        f"contained by the current block {start}:{top}."
                    )
                    if raise_error:
                        raise ValueError(err_msg)
                    warnings.warn(f"{err_msg} No scope added to test split.")
                    train_scopes.append((ts_id, slice(start, stop)))
                else:
                    test_scopes.append((ts_id, slice(test_start, stop)))
                    train_scopes.append((ts_id, slice(start, new_start)))
        train_corpus = TimeSeriesCorpus(
            self.instances,
            train_scopes,
            categorical_encoders=self.categorical_encoders,
            numerical_encoders=self.numerical_scalers,
            add_series_id=self.add_series_id,
        )
        test_corpus = TimeSeriesCorpus(
            self.instances,
            test_scopes,
            categorical_encoders=self.categorical_encoders,
            numerical_encoders=self.numerical_scalers,
            add_series_id=self.add_series_id,
        )
        return train_corpus, test_corpus


if __name__ == "__main__":
    instances = []
    for i in range(100):
        start_time = pd.Timestamp("2011-1-1")
        time_unit = "d"
        data = np.random.randn(100, 2)
        target_names = "data"
        t = TimeSeries(
            data,
            series_name=f"instance_{i+1}",
            start_time=start_time,
            time_unit=time_unit,
            target_names=target_names,
        )
        t.add_features("f1", np.random.randn(100, 1), "num", "revealed")
        t.add_features("f2", np.random.randn(100, 1), "num", "observed")
        t.add_features(
            "f3", np.random.randint(100, 200, (100,)), "cat", "revealed"
        )
        t.add_features(
            "f4", np.random.randint(1000, 1200, (100,)), "cat", "observed"
        )
        t.add_features("f5", np.array(i + 1), "cat", "static")
        t.add_features("f6", np.random.randn(2) + 12, "num", "static")
        instances.append(t)
    c = TimeSeriesCorpus(instances, add_series_id=True)

    t1, t2 = c.split_by_timestamp(pd.Timestamp("2011-04-01"), 4, 10)
    t1, t2 = c.split_from_end(4, 10)
    t1, t2 = c.split_from_end_by_blocks(4, 10, 4, 4)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from multivariate.datasets.dataset import (
    electricity,
    exchange_rate,
    solar,
    traffic,
    wiki,
    taxi_30min,
)

DATASETS = ["electricity", "exchange_rate", "solar", "traffic", "wiki", "taxi"]


def get_dataset(dataset, max_target_dim):
    if dataset == "electricity":
        return electricity(max_target_dim=max_target_dim)
    elif dataset == "exchange_rate":
        return exchange_rate(max_target_dim=max_target_dim)
    elif dataset == "solar":
        return solar(max_target_dim=max_target_dim)
    elif dataset == "traffic":
        return traffic(max_target_dim=max_target_dim)
    elif dataset == "wiki":
        return wiki(max_target_dim=max_target_dim)
    elif dataset == "taxi":
        return taxi_30min(max_target_dim=max_target_dim)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List
import torch
from lightkit.nn import Configurable
from dataclasses import dataclass, field
from meta.models.registry import register_model

from meta.data.batch import SeriesBatch
from meta.models.model import SeriesModel

from .components import (
    IdentityFeatureExtractor,
    LSTMQueryEncoder,
    FeedForwardQuantileDecoder,
)


@dataclass
class LSTMEncoderFeedforwardDecoderConfig:
    """
    Configuration class for a LSTMEncoderFeedforwardDecoder.

    See also:
        :class:`LSTMEncoderFeedforwardDecoder`
    """

    # The length of the prediction horizon
    prediction_length: int = 1
    # The quantiles the model predicts
    quantiles: List[str] = field(
        default_factory=lambda: [
            "0.02",
            "0.1",
            "0.25",
            "0.5",
            "0.75",
            "0.9",
            "0.98",
        ]
    )
    query_out_channels: int = 64
    # The number of layers in the encoding LSTM
    num_lstm_layers: int = 1
    decoder_hidden_size: int = 32


@register_model
class LSTMEncoderFeedforwardDecoder(
    SeriesModel, Configurable[LSTMEncoderFeedforwardDecoderConfig]
):
    """
    A simple base model that does not make use of a support set.
    """

    def __init__(self, config: LSTMEncoderFeedforwardDecoderConfig):
        Configurable.__init__(self, config)
        SeriesModel.__init__(self)
        self.feature_extractor = IdentityFeatureExtractor()
        self.query_enc = LSTMQueryEncoder(
            input_size=1,
            num_layers=config.num_lstm_layers,
            hidden_size=config.query_out_channels,
        )
        self.decoder = FeedForwardQuantileDecoder(
            embed_size=0,
            q_size=config.query_out_channels,
            hidden_size=config.decoder_hidden_size,
            prediction_length=config.prediction_length,
            num_quantiles=len(config.quantiles),
        )

    @classmethod
    def name(cls) -> str:
        return "lstm_feedforward"

    def forward(self, supps: SeriesBatch, query: SeriesBatch) -> torch.Tensor:
        """
        Computes the forecasts from the provided queries.

        Parameters
        ----------
        supps: Not used!
        query: The query.
        """

        x = self.feature_extractor(query)
        x = self.query_enc(query=x)
        x = self.decoder(query=x)
        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union, Tuple, List, Iterator
from pathlib import Path
from functools import partial
from copy import deepcopy
import math

import torch as pt
from torch import nn
from torch.nn.parallel import DistributedDataParallel
from tqdm import tqdm

from .callback import Callback, CallbackList, Optimization, DebugStopper
from .distributed import is_distributed, synchronize, is_main_process

if TYPE_CHECKING:
    from ..dataset import MetaDataset
    from ..metrics import MeterDict


class Trainer(object):
    def __init__(
        self,
        dataset: MetaDataset,
        model: nn.Module,
        optimizer: partial,
        scheduler: Optional[partial],
        metrics: MeterDict,
        callbacks: List[Callback],
        log_dir: Union[Path, str],
        n_epochs: int,
        nb_epoch: Optional[int],
        batch_size: Union[int, Tuple[int, int]],
        n_loader_workers: int = 0,
        cuda_device: int = 0,
        debug: bool = False,
    ) -> None:
        self.dataset = dataset
        self.cuda_device = cuda_device if pt.cuda.is_available() else -1
        if self.cuda_device >= 0:
            pt.cuda.set_device(self.cuda_device)
            self.model = model.cuda(self.cuda_device)
        else:
            self.model = model.cpu()
        if is_distributed():
            device_ids = None if self.cuda_device < 0 else [self.cuda_device]
            self.model = DistributedDataParallel(
                self.model, device_ids=device_ids
            )
        self.optimizer = optimizer(self.model.parameters())
        if scheduler is None:
            self.scheduler = None
        else:
            self.scheduler = scheduler(self.optimizer)
        self.metrics = metrics
        self.register_callback(*callbacks)
        self.log_dir = Path(log_dir) if isinstance(log_dir, str) else log_dir
        self.n_epochs = n_epochs
        self.nb_epoch = nb_epoch
        if isinstance(batch_size, int):
            self.batch_size = self.eval_batch_size = batch_size
        else:
            self.batch_size, self.eval_batch_size = batch_size
        self.n_loader_workers = n_loader_workers
        self.debug = debug

        self.global_step = 0
        self._signal_exit = False
        self._signal_break = False
        self._signal_valid_break = False

    def register_callback(self, *callbacks: Callback):
        if hasattr(self, "callbacks"):
            for callback in callbacks:
                self.callbacks.append(callback)
        else:
            callbacks = list(callbacks)
            has_optimization = False
            has_debug_stopper = False
            for callback in callbacks:
                if isinstance(callback, Optimization):
                    has_optimization = True
                elif isinstance(callback, DebugStopper):
                    has_debug_stopper = True
            if not has_optimization:
                callbacks.insert(0, Optimization())
            if not has_debug_stopper:
                callbacks.insert(0, DebugStopper())
            self.callbacks = CallbackList(*callbacks, trainer=self)

    def get_model_core(self) -> nn.Module:
        if is_distributed():
            return self.model.module
        else:
            return self.model

    def dump(
        self,
        tag: str,
        save_model: bool = True,
        save_optimizer: bool = True,
        save_scheduler: bool = True,
        save_metrics: bool = True,
    ) -> None:
        if is_main_process():
            # in distributed training, only model from process 0 is dumped
            state = {"global_step": self.global_step}
            if save_model:
                state["model"] = self.get_model_core().state_dict()
            if save_optimizer:
                state["optimizer"] = self.optimizer.state_dict()
            if save_scheduler and (self.scheduler is not None):
                state["scheduler"] = self.scheduler.state_dict()
            if save_metrics:
                state["metrics"] = self.metrics.state_dict()
            pt.save(state, self.log_dir.joinpath(f"{tag}.pt.tar"))
        # make sure non-base processes wait until process 0 finish checkpointing
        synchronize()

    def load(
        self,
        tag: str,
        load_model: bool = True,
        load_optimizer: bool = True,
        load_scheduler: bool = True,
        load_metrics: bool = True,
    ) -> None:
        try:
            path = self.log_dir.joinpath(f"{tag}.pt.tar")
            state = pt.load(
                path,
                map_location=(
                    f"cuda:{self.cuda_device}"
                    if self.cuda_device >= 0
                    else "cpu"
                ),
            )
            print(f"Load checkpoint from {path}")
        except FileNotFoundError:
            raise ValueError(f"invalid tag {tag}")
        err_msg = "the checkpoint does not have {} data."
        global_step = state.get("global_step")
        if global_step is None:
            raise KeyError(err_msg.format("global_step"))
        self.global_step = global_step
        if load_model:
            module_state = state.get("model")
            if module_state is None:
                raise KeyError(err_msg.format("model"))
            self.get_model_core().load_state_dict(module_state)
        if load_optimizer:
            optimizer_state = state.get("optimizer")
            if optimizer_state is None:
                raise KeyError(err_msg.format("optimization"))
            self.optimizer.load_state_dict(optimizer_state)
        if load_scheduler and self.scheduler is not None:
            scheduler_state = state.get("scheduler")
            if scheduler_state is None:
                raise KeyError(err_msg.format("scheduler"))
            self.scheduler.load_state_dict(scheduler_state)
        if load_metrics:
            metrics_state = state.get("metrics")
            if metrics_state is None:
                raise KeyError(err_msg.format("metrics"))
            self.metrics.load_state_dict(metrics_state)

    @property
    def train_loader(self) -> Iterator:
        return self.dataset.train_loader(
            batch_size=self.batch_size,
            shuffle=(not self.debug),
            cuda_device=self.cuda_device,
            n_workers=self.n_loader_workers,
            n_batches=self.nb_epoch,
        )

    @property
    def valid_loader(self) -> Iterator:
        return self.dataset.valid_loader(
            batch_size=self.eval_batch_size,
            cuda_device=self.cuda_device,
            n_workers=self.n_loader_workers,
        )

    def fit(self) -> None:
        self.callbacks.on_train_begin()
        try:
            self._epoch_loop()
        except KeyboardInterrupt:
            if not self.debug:
                if input("Do you want to save the current model? (Y/n)") in [
                    "Y",
                    "y",
                ]:
                    self.dump(tag=f"epoch{epoch+1:02d}_interrupt")
            return
        finally:
            self.callbacks.on_train_end()

    def _epoch_loop(self) -> None:
        for epoch in range(self.n_epochs):
            self.model.train()
            self.callbacks.on_epoch_begin(epoch)
            self._batch_loop(epoch)
            self.evaluate(epoch)
            self.callbacks.on_epoch_end(epoch)
            self.metrics.restart()
            if self._signal_exit:
                break

    def _batch_loop(self, epoch: int) -> None:
        for batch, data in enumerate(self.train_loader):
            self.callbacks.on_batch_begin(batch)
            self._train(*data)
            self.callbacks.on_batch_end(batch)
            if self._signal_break:
                break

    def _train(self, *data):
        raise NotImplementedError

    def evaluate(self, epoch: int) -> None:
        _ = self.model.eval()
        with pt.no_grad():
            self.callbacks.on_valid_begin(epoch)
            self._valid_batch_loop(epoch)
            self.callbacks.on_valid_end(epoch)
        synchronize()

    def _valid_batch_loop(self, epoch: int) -> None:
        for batch, data in enumerate(self.valid_loader):
            self.callbacks.on_valid_batch_begin(batch)
            self._validate(*data)
            self.callbacks.on_valid_batch_end(batch)
            if self._signal_valid_break:
                break

    def _validate(self, *data):
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from typing import List

from mxnet.gluon import HybridBlock, nn

from src.gluonts.core.component import validated
from src.gluonts.dataset.common import Dataset
from src.gluonts.dataset.field_names import FieldName
from src.gluonts.dataset.loader import (
    DataLoader,
    TrainDataLoader,
    ValidationDataLoader,
)
from src.gluonts.model.predictor import Predictor
from src.gluonts.mx.batchify import batchify
from src.gluonts.mx.block.feature import FeatureEmbedder
from src.gluonts.mx.block.rnn import RNN
from src.gluonts.mx.distribution import DistributionOutput, StudentTOutput
from src.gluonts.mx.model.estimator import GluonEstimator
from src.gluonts.mx.model.predictor import RepresentableBlockPredictor
from src.gluonts.mx.trainer import Trainer
from src.gluonts.mx.util import get_hybrid_forward_input_names
from src.gluonts.time_feature import time_features_from_frequency_str
from src.gluonts.transform import (
    AddTimeFeatures,
    AsNumpyArray,
    InstanceSplitter,
    SelectFields,
    SetFieldIfNotPresent,
    TestSplitSampler,
    Transformation,
    ValidationSplitSampler,
)

from ._network import CanonicalPredictionNetwork, CanonicalTrainingNetwork


class CanonicalEstimator(GluonEstimator):
    @validated()
    def __init__(
        self,
        model: HybridBlock,
        is_sequential: bool,
        freq: str,
        context_length: int,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        num_parallel_samples: int = 100,
        cardinality: List[int] = list([1]),
        embedding_dimension: int = 10,
        distr_output: DistributionOutput = StudentTOutput(),
        batch_size: int = 32,
    ) -> None:
        super().__init__(trainer=trainer, batch_size=batch_size)

        # TODO: error checking
        self.freq = freq
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.distr_output = distr_output
        self.num_parallel_samples = num_parallel_samples
        self.cardinality = cardinality
        self.embedding_dimensions = [embedding_dimension for _ in cardinality]
        self.model = model
        self.is_sequential = is_sequential

    def create_transformation(self) -> Transformation:
        return (
            AsNumpyArray(field=FieldName.TARGET, expected_ndim=1)
            + AddTimeFeatures(
                start_field=FieldName.START,
                target_field=FieldName.TARGET,
                output_field=FieldName.FEAT_TIME,
                time_features=time_features_from_frequency_str(self.freq),
                pred_length=self.prediction_length,
            )
            + SetFieldIfNotPresent(
                field=FieldName.FEAT_STATIC_CAT, value=[0.0]
            )
            + AsNumpyArray(field=FieldName.FEAT_STATIC_CAT, expected_ndim=1)
        )

    def _create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        instance_sampler = {
            "training": ValidationSplitSampler(
                min_future=self.prediction_length
            ),
            "validation": ValidationSplitSampler(
                min_future=self.prediction_length
            ),
            "test": TestSplitSampler(),
        }[mode]

        return InstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=instance_sampler,
            time_series_fields=[FieldName.FEAT_TIME],
            past_length=self.context_length,
            future_length=self.prediction_length,
        )

    def create_training_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(CanonicalTrainingNetwork)
        instance_splitter = self._create_instance_splitter("training")
        return TrainDataLoader(
            dataset=data,
            transform=instance_splitter + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.trainer.ctx, dtype=self.dtype),
            **kwargs,
        )

    def create_validation_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(CanonicalTrainingNetwork)
        instance_splitter = self._create_instance_splitter("validation")
        return ValidationDataLoader(
            dataset=data,
            transform=instance_splitter + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.trainer.ctx, dtype=self.dtype),
        )

    def create_training_network(self) -> CanonicalTrainingNetwork:
        return CanonicalTrainingNetwork(
            embedder=FeatureEmbedder(
                cardinalities=self.cardinality,
                embedding_dims=self.embedding_dimensions,
            ),
            model=self.model,
            distr_output=self.distr_output,
            is_sequential=self.is_sequential,
        )

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: CanonicalTrainingNetwork,
    ) -> Predictor:
        prediction_splitter = self._create_instance_splitter("test")

        prediction_net = CanonicalPredictionNetwork(
            embedder=trained_network.embedder,
            model=trained_network.model,
            distr_output=trained_network.distr_output,
            is_sequential=trained_network.is_sequential,
            prediction_len=self.prediction_length,
            num_parallel_samples=self.num_parallel_samples,
            params=trained_network.collect_params(),
        )

        return RepresentableBlockPredictor(
            input_transform=transformation + prediction_splitter,
            prediction_net=prediction_net,
            batch_size=self.batch_size,
            prediction_length=self.prediction_length,
            ctx=self.trainer.ctx,
        )


class CanonicalRNNEstimator(CanonicalEstimator):
    @validated()
    def __init__(
        self,
        freq: str,
        context_length: int,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        num_layers: int = 1,
        num_cells: int = 50,
        cell_type: str = "lstm",
        num_parallel_samples: int = 100,
        cardinality: List[int] = list([1]),
        embedding_dimension: int = 10,
        distr_output: DistributionOutput = StudentTOutput(),
    ) -> None:
        model = RNN(
            mode=cell_type, num_layers=num_layers, num_hidden=num_cells
        )

        super().__init__(
            model=model,
            is_sequential=True,
            freq=freq,
            context_length=context_length,
            prediction_length=prediction_length,
            trainer=trainer,
            num_parallel_samples=num_parallel_samples,
            cardinality=cardinality,
            embedding_dimension=embedding_dimension,
            distr_output=distr_output,
        )


class MLPForecasterEstimator(CanonicalEstimator):
    @validated()
    def __init__(
        self,
        freq: str,
        context_length: int,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        hidden_dim_sequence=list([50]),
        num_parallel_samples: int = 100,
        cardinality: List[int] = list([1]),
        embedding_dimension: int = 10,
        distr_output: DistributionOutput = StudentTOutput(),
    ) -> None:
        model = nn.HybridSequential()

        for layer, layer_dim in enumerate(hidden_dim_sequence):
            model.add(
                nn.Dense(
                    layer_dim,
                    flatten=False,
                    activation="relu",
                    prefix="mlp_%d_" % layer,
                )
            )

        super().__init__(
            model=model,
            is_sequential=False,
            freq=freq,
            context_length=context_length,
            prediction_length=prediction_length,
            trainer=trainer,
            num_parallel_samples=num_parallel_samples,
            cardinality=cardinality,
            embedding_dimension=embedding_dimension,
            distr_output=distr_output,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Union, List
from pytorch_lightning import Callback, Trainer
import matplotlib.pyplot as plt

from meta.data.batch import TripletBatch
from meta.models.module import MetaLightningModule
from .common import get_save_dir_from_csvlogger, get_loss_steps
from meta.data.batch import SeriesBatch
from meta.vis.forecast import (
    plot_quantile_forecast,
    plot_forecast_supportset_attention,
)


class ForecastPlotLoggerCallback(Callback):
    """
    A callback that stores plots of the  predictions for a collection of
    samples every n epochs. The plots display the query (past and future) and
    forecasted quantiles. This callback is intended for models without support
    set and attention mechanism.

    Args:
        log_batch: For each sample in the batch the prediction is plotted when the callback is called.
            The batch size should not be too large (i.e. < 10).
        quantiles: The quantiles that are predicted.
        split: Specifies the split the batch comes from (i.e. from training or validation split)
        every_n_epochs: Specifies how often the plots are generated.
            Setting this to a large value can save time since plotting can be time consuming
            (especially when small datasets are used).
    """

    def __init__(
        self,
        log_batch: TripletBatch,
        quantiles: List[str],
        split: Optional[Union["train", "val"]] = None,
        every_n_epochs: int = 1,
    ):
        super().__init__()
        self.log_batch = log_batch
        self.quantiles = quantiles
        self.split = split
        self.every_n_epochs = every_n_epochs

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if trainer.current_epoch % self.every_n_epochs:
            return
        # Bring the tensors to CPU
        log_batch = self.log_batch.to(device=pl_module.device)
        query_past = log_batch.query_past
        query_future = log_batch.query_future
        support_set = log_batch.support_set

        # Get model prediction
        pred = pl_module.model(supps=support_set, query=query_past)
        pred = SeriesBatch(
            pred, query_future.lengths, query_future.split_sections
        ).unpad(to_numpy=True, squeeze=True)
        qp = query_past.unpad(to_numpy=True, squeeze=True)
        qf = query_future.unpad(to_numpy=True, squeeze=True)
        fig = plot_quantile_forecast(qp, qf, pred, quantiles=self.quantiles)
        fig.savefig(
            get_save_dir_from_csvlogger(trainer.logger)
            / f"pred_{self.split + '_' if self.split else ''}ep{trainer.current_epoch}.png",
            bbox_inches="tight",
        )
        plt.close(fig)


class ForecastSupportSetAttentionPlotLoggerCallback(Callback):
    """
    A callback that stores plots of the  predictions for a collection of
    samples every n epochs. The plots display the query (past and future),
    forecasted quantiles and the time series in the support set of this sample
    aligned with their attention scores. This callback works only for models
    with attention mechanism!

    Args:
        log_batch: For each sample in the batch the prediction is plotted when the callback is called.
            The batch size should not be too large (i.e. < 10).
        quantiles: The quantiles that are predicted.
        split: Specifies the split the batch comes from (i.e. from training or validation split)
        every_n_epochs: Specifies how often the plots are generated.
            Setting this to a large value can save time since plotting can be time consuming
            (especially when small datasets are used).
    """

    def __init__(
        self,
        log_batch: TripletBatch,
        quantiles: List[str],
        split: Optional[Union["train", "val"]] = None,
        every_n_epochs: int = 1,
    ):
        super().__init__()
        self.log_batch = log_batch
        self.quantiles = quantiles
        self.split = split
        self.every_n_epochs = every_n_epochs

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if trainer.current_epoch % self.every_n_epochs:
            return
        # Bring the tensors to CPU
        log_batch = self.log_batch.to(device=pl_module.device)
        query_past = log_batch.query_past
        query_future = log_batch.query_future
        support_set = log_batch.support_set

        # Get model prediction
        pred, attention = pl_module.model(
            supps=support_set,
            query=query_past,
            return_attention=True,
        )
        pred = SeriesBatch(
            pred, query_future.lengths, query_future.split_sections
        ).unpad(to_numpy=True, squeeze=True)
        qp = query_past.unpad(to_numpy=True, squeeze=True)
        qf = query_future.unpad(to_numpy=True, squeeze=True)
        supps = support_set.unpad(to_numpy=True, squeeze=False)
        n_supps = support_set.split_sections[0]
        attention = attention.reshape(n_supps * attention.size()[0], 1, -1)
        attention = SeriesBatch(
            attention.transpose(1, 2),
            support_set.lengths,
            support_set.split_sections,
        ).unpad(to_numpy=True, squeeze=False)
        # attention = [[tensor_to_np(att) for att in supps] for supps in attention]

        fig = plot_forecast_supportset_attention(
            qp,
            qf,
            pred,
            supps=supps,
            attention=attention,
            quantiles=self.quantiles,
        )
        fig.savefig(
            get_save_dir_from_csvlogger(trainer.logger)
            / f"pred_supps_{self.split + '_' if self.split else ''}ep{trainer.current_epoch}.png",
            bbox_inches="tight",
        )
        plt.close(fig)


class LossPlotLoggerCallback(Callback):
    """
    A callback that stores plots of the training and macro-averaged validation
    loss curve every n epochs.

    Args:
        every_n_epochs: Specifies how often the plots are generated.
            Setting this to a large value can save time since plotting can be time consuming
            (especially when small datasets are used).
    """

    def __init__(self, every_n_epochs: int = 1):
        super().__init__()
        self.every_n_epochs = every_n_epochs

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if (trainer.current_epoch < 1) or (
            trainer.current_epoch % self.every_n_epochs
        ):
            return
        self.plot_loss(trainer=trainer, pl_module=pl_module)

    def on_fit_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if trainer.current_epoch < 1:
            return
        self.plot_loss(trainer=trainer, pl_module=pl_module)

    def plot_loss(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        train_loss, train_steps = get_loss_steps("train_loss", trainer)
        val_loss_macro, val_steps = get_loss_steps("val_loss_macro", trainer)

        fig, ax = plt.subplots()
        ax.semilogy(train_steps, train_loss, label="train_loss", alpha=0.5)
        ax.semilogy(
            val_steps, val_loss_macro, label="val_loss_macro", alpha=0.5
        )
        ax2 = ax.twinx()
        ax2.set_ylabel("crps scores", color="red")
        if pl_module.val_crps_scaled_macro:
            val_crps_scaled_macro, _ = get_loss_steps(
                "val_crps_scaled_macro", trainer
            )
            ax2.semilogy(
                val_steps,
                val_crps_scaled_macro,
                "r--",
                label="val_crps_scaled_macro",
            )
        else:
            val_crps_macro, _ = get_loss_steps("val_crps_macro", trainer)
            ax2.semilogy(
                val_steps, val_crps_macro, "r--", label="val_crps_macro"
            )
        ax2.legend(loc="upper right")

        ticks = range(0, trainer.current_epoch + 1, self.every_n_epochs)
        labels = [str(t) for t in ticks]
        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)
        ax.legend(loc="upper left")

        fig.savefig(
            get_save_dir_from_csvlogger(trainer.logger) / f"loss.png",
            bbox_inches="tight",
        )
        plt.close(fig)


class CheatLossPlotLoggerCallback(LossPlotLoggerCallback):
    """
    A callback that stores plots of the training and multiple validation losses
    curve every n epochs.

    Args:
        every_n_epochs: Specifies how often the plots are generated.
            Setting this to a large value can save time since plotting can be time consuming
            (especially when small datasets are used).
        dataset_names_val: The names of the datasets that the validation loss should be plotted for.
    """

    def __init__(self, dataset_names_val, **kwargs):
        super().__init__(**kwargs)
        self.dataset_names_val = dataset_names_val

    def plot_loss(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        train_loss, train_steps = get_loss_steps("train_loss", trainer)

        fig, ax = plt.subplots()
        ax.semilogy(train_steps, train_loss, label="train_loss", alpha=0.5)
        for d_name in self.dataset_names_val:
            val_loss, val_steps = get_loss_steps(f"{d_name}_val_loss", trainer)
            ax.semilogy(val_steps, val_loss, label=f"{d_name}_val_loss")

        ticks = range(0, trainer.current_epoch + 1, self.every_n_epochs)
        labels = [str(t) for t in ticks]
        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)
        ax.legend(loc="upper left")

        fig.savefig(
            get_save_dir_from_csvlogger(trainer.logger) / f"loss.png",
            bbox_inches="tight",
        )
        plt.close(fig)


class MacroCRPSPlotCallback(Callback):
    """
    A callback that stores plots of the validation losses and a macro-averaged
    validation loss curve every n epochs.

    Args:
        every_n_epochs: Specifies how often the plots are generated.
            Setting this to a large value can save time since plotting can be time consuming
            (especially when small datasets are used).
    """

    def __init__(
        self,
        every_n_epochs: int = 1,
    ):
        super().__init__()
        self.every_n_epochs = every_n_epochs
        # We only compute this on the data as it is (no rescaling)

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if trainer.current_epoch % self.every_n_epochs:
            return
        self.plot_crps(trainer, scaled=False)
        dm_super = trainer.lightning_module.trainer.datamodule
        if dm_super.standardize:
            self.plot_crps(trainer, scaled=True)

    def on_fit_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if trainer.current_epoch < 1:
            return
        self.plot_crps(trainer, scaled=False)
        dm_super = trainer.lightning_module.trainer.datamodule
        if dm_super.standardize:
            self.plot_crps(trainer, scaled=True)

    def plot_crps(self, trainer: Trainer, scaled: bool) -> None:
        cm = plt.get_cmap("tab20")
        suffix = "_scaled" if scaled else ""
        dm_super = trainer.lightning_module.trainer.datamodule

        fig, ax = plt.subplots()
        for i, dm_val in enumerate(dm_super.data_modules_val):
            c, steps = get_loss_steps(
                f"{dm_val.dataset_name}_val_crps{suffix}", trainer
            )
            ax.semilogy(
                steps,
                c,
                label=dm_val.dataset_name,
                alpha=1.0,
                color=cm.colors[i],
            )

        macro, steps = get_loss_steps(f"val_crps{suffix}_macro", trainer)
        ax.semilogy(steps, macro, "r--", label=f"val_crps{suffix}_macro")
        ticks = range(0, trainer.current_epoch + 1, self.every_n_epochs)
        labels = [str(t) for t in ticks]
        ax.set_xticks(ticks)
        ax.set_xticklabels(labels, rotation=45)
        ax.legend(
            ncol=2,
            loc="upper center",
            handletextpad=1.5,
            bbox_to_anchor=(0.5, 1.3),
            framealpha=0.5,
            prop={"size": 8},
        )

        fig.savefig(
            get_save_dir_from_csvlogger(trainer.logger)
            / f"val_crps{suffix}_macro.png",
            bbox_inches="tight",
        )
        plt.close(fig)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pytorch_lightning import Callback, Trainer, LightningModule


class ParameterCountCallback(Callback):  # type: ignore
    """
    This callback allows counting model parameters during training.

    The output is printed to the console and can be retrieved from the log
    files.
    """

    def __init__(self) -> None:
        super().__init__()

    def on_pretrain_routine_start(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        # compute number of model params
        model_total_params = sum(
            p.numel() for p in pl_module.model.parameters()
        )
        model_total_trainable_params = sum(
            p.numel() for p in pl_module.model.parameters() if p.requires_grad
        )

        # log
        print("\n" + f"model_total_params: {model_total_params},")
        print(
            "\n"
            + f"model_total_trainable_params: {model_total_trainable_params},"
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
import torch
from torch import nn
from meta.data.batch import SeriesBatch


class MetaModel(ABC, nn.Module):
    """
    Base class for all meta models that make predictions based on a support set
    and queries.
    """

    @abstractmethod
    def forward(self, supps: SeriesBatch, query: SeriesBatch) -> torch.Tensor:
        """
        Computes the forecasts for query past from the support set.

        Parameters
        ----------
        supps
            The support set.
        query
            The queries to be forecasted consisting.

        Returns
        -------
        torch.Tensor
            The forecasts for the entire prediction length with shape `[num_queries,
            prediction_length, *]`.
        """

    @property
    def device(self):
        return next(self.parameters()).device


class SeriesModel(ABC, nn.Module):
    """
    Base class for all models that make predictions based on a query only.
    """

    @abstractmethod
    def forward(self, query: SeriesBatch) -> torch.Tensor:
        """
        Computes the forecasts from query past.

        Parameters
        ----------
        query
            The query to be forecasted.

        Returns
        -------
        torch.Tensor
            The forecasts for the entire prediction length with shape `[num_queries,
            prediction_length, *]`.
        """

    @property
    def device(self):
        return next(self.parameters()).device



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional

import numpy as np


def fill_forward(a: np.ndarray, fill_start=None) -> np.ndarray:
    """
    Forward fill an array.

    If `fill_start` is not None, then the NaNs in the beginning of the array
    will be filled with `fill_start`.
    """
    a = np.array(a)
    assert a.ndim == 1

    # forward fill labels
    idx = np.where(~np.isnan(a), np.arange(a.shape[0]), 0)
    np.maximum.accumulate(idx, out=idx)
    a = a[idx]

    # nan values at the beginning of window are left
    if fill_start is not None:
        a[~np.isfinite(a)] = fill_start

    return a


def labels_filter(
    a: np.ndarray,
    num_open: int,
    num_clear: int,
    initial_label: int = 0,
    forward_fill: bool = False,
):
    if forward_fill:
        a = fill_forward(a)

    state = initial_label
    count_high = 0
    count_low = 0
    output = []
    for label in a:
        if label and not np.isnan(label):
            count_low = 0
            count_high += 1
        else:
            count_high = 0
            count_low += 1
        if state and count_low == num_clear:
            state = 0
        if not state and count_high == num_open:
            state = 1
        output.append(state)
    return np.array(output)


def n_k_filter(
    a: np.ndarray,
    num_open: int,
    num_clear: int,
    num_open_suff: Optional[int] = None,
    num_clear_suff: Optional[int] = None,
    initial_label: int = 0,
    forward_fill: bool = False,
) -> np.ndarray:
    """
    Implements the (N, K)-filter that opens/closes anomalies based on observing
    K or greater labels in the last N time steps.

    Parameters
    ----------
    a
        input array, type boolean
    num_open
        number of time steps to look back for opening an anomaly
    num_clear
        number of time steps to look back for clearing an anomaly
    num_open_suff
        number of positive labels in the lookback period sufficient
        to open an anomaly. defaults to num_open.
    num_clear_suff
        number of positive labels in the lookback period sufficient
        to clear an anomaly. defaults to num_clear.
    initial_label
        1 if the initial state is an anomaly, 0 if not. default 0.
    """
    num_open_suff = num_open_suff or num_open
    num_clear_suff = num_clear_suff or num_clear

    if forward_fill:
        a = fill_forward(a)

    def causal_conv(a, f):
        if len(f) > 1:
            return np.convolve(a, f)[: -(len(f) - 1)]
        return np.convolve(a, f)

    up = causal_conv(a > 0, np.ones(num_open)) >= num_open_suff
    down = causal_conv(a == 0, np.ones(num_clear)) >= num_clear_suff

    s = initial_label
    out = []
    for i in range(len(up)):
        s = up[i] if s == 0 else ~down[i]
        out.append(s)
    return np.array(out, dtype=int)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
import torch
from torch import nn
import torch.nn.functional as F


class Decoder(nn.Module, ABC):
    """
    Base class for decoders.
    """

    @abstractmethod
    def forward(
        self, query: torch.Tensor, value: torch.Tensor
    ) -> torch.Tensor:
        """
        Forecast from encoded (query, support set) (typically via attention)
        and encoded queries.

        Parameters
        ----------
        value: Encoded (query, support set).
        query: Encoded query.

        Returns
        -------
        torch.Tensor: predictions of size [n_q, *, *]
        """


class FeedForwardQuantileDecoder(Decoder):
    """
    A simple linear decoder.

    Parameters
    ----------
    embed_size: The size of the value embedding dimension.
    q_size: The size of the query encoding.
    prediction_length: The length of the prediction horizon.
    num_quantiles: The number of quantiles to be predicted.


    Returns
    -------
    torch.Tensor: predictions of size [batch, prediction_length, num_quantiles]
    """

    def __init__(
        self,
        num_quantiles: int,
        embed_size: int = 10,
        q_size: int = 20,
        hidden_size: int = 32,
        prediction_length: int = 1,
    ):
        super().__init__()
        self.num_quantiles = num_quantiles
        self.prediction_length = prediction_length
        self.fc1 = nn.Linear(embed_size + q_size, hidden_size)
        self.fc2 = nn.Linear(
            hidden_size, prediction_length * self.num_quantiles
        )

    def forward(
        self, query: torch.Tensor, value: torch.Tensor = None
    ) -> torch.Tensor:
        if value is not None:
            x = torch.cat((value, query.sequences), dim=1)
        else:
            x = query.sequences
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x)
        x = x.reshape(-1, self.prediction_length, self.num_quantiles)
        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Union, List
from pytorch_lightning import Callback, Trainer
import torch
from src.gluonts.time_feature import get_seasonality

from meta.models.module import MetaLightningModule
from meta.common.torch import tensor_to_np
from meta.metrics.numpy import compute_metrics


class QuantileMetricLoggerCallback(Callback):
    """
    A callback that computes additional metrics on a numpy representation of
    the dataset every n epochs. The computed values are logged to the output
    file of the pytorch lightning logger.

    Args:
        quantiles: The quantiles that are predicted.
        split: Specifies the split the batch comes from (i.e. from training or validation split)
        every_n_epochs: Specifies how often the plots are generated.
            Setting this to a large value can save time since plotting can be time consuming
            (especially when small datasets are used).
    """

    def __init__(
        self,
        quantiles: List[str],
        split: Optional[Union["train", "val"]] = None,
        every_n_epochs: int = 1,
    ):
        super().__init__()
        self.quantiles = quantiles
        self.split = split
        self.every_n_epochs = every_n_epochs

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: MetaLightningModule
    ) -> None:
        if trainer.current_epoch % self.every_n_epochs:
            return
        dm_super = trainer.lightning_module.trainer.datamodule
        dm_val = dm_super.data_modules_val[0]
        dl = dm_val.val_dataloader()
        split = dm_val.splits.val()
        pred = []
        for batch in dl:
            batch = batch.to(pl_module.device)
            pred.append(
                pl_module.model(
                    supps=batch.support_set, query=batch.query_past
                )
            )

        # redo standardization for evaluation
        pred = split.data().rescale_dataset(torch.cat(pred, dim=0).cpu())
        pred = tensor_to_np(pred)

        # use only the length that should be included for evaluation
        pred = pred[:, : dm_val.prediction_length, ...]
        m = compute_metrics(
            pred,
            split.evaluation(),
            quantiles=self.quantiles,
            seasonality=get_seasonality(dm_val.meta.freq),
        )
        self.log("metrics", m)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List
import numpy as np
import matplotlib.pyplot as plt


def plot_forecast_supportset_attention(
    query_past: List[np.ndarray],
    query_future: List[np.ndarray],
    pred: List[np.ndarray],
    supps: List[np.ndarray],
    attention: List[np.ndarray],
    quantiles: List[str],
) -> plt.Figure:
    """
    Plots the provided forecasts for each sample with confidence intervals
    using all provided quantiles. Furthermore, plots the time series in the
    support set of this sample aligned with their attention scores.

    Parameters
    ----------
    query_past: List[np.ndarray] of length n_samples, each array has shape [context length]
        The past query which the model uses to make a prediction.
    query_future: List[np.ndarray] of length n_samples, each array has shape [prediction horizon length]
        The ground truth for the forecast.
    pred: List[np.ndarray], of length n_samples, each array has shape  [prediction horizon length, n_quantiles]
        The prediction of the model.
    supps: List[np.ndarray], length is n_samples, each item has shape[supps_size, support ts length]
        The support sets for each query.
    attention: List[np.ndarray], length is n_samples, each item has shape[supps_size, support ts length]
        The attention scores for each support set times series.
    quantiles: List[float], shape[n_quantiles,]

    Returns
    -------
    plt.figure.Figure
        A plot containing one subplot for each sample. Each subplot displays the ground truth, predicted confidence
        intervals, the predicted median. Furthermore, the time series in the support set aligned with their attention
        scores.
    """
    supps_size = len(supps[0])
    n_samples = len(query_past)
    plots_per_sample = supps_size + 1
    n_plots = plots_per_sample * n_samples
    fig, subs = plt.subplots(n_plots, figsize=(10, 4 * n_plots))
    for i in range(n_samples):
        # plot prediction
        subs[plots_per_sample * i].set_title(f"query {i}")
        _plot_quantile_forecast(
            subs[plots_per_sample * i],
            query_past[i],
            query_future[i],
            pred[i],
            quantiles,
        )
        # plot support ts and attention weights
        for j in range(1, supps_size + 1):
            subs[plots_per_sample * i + j].set_title(
                f"support ts {j} for query {i}"
            )
            subs[plots_per_sample * i + j].plot(supps[i][j - 1].squeeze())

            subs[plots_per_sample * i + j].sharex(subs[plots_per_sample * i])
            subs[plots_per_sample * i + j].sharey(subs[plots_per_sample * i])
            # twin object for two different y-axis on the sample plot
            ax2 = subs[plots_per_sample * i + j].twinx()
            for head in range(attention[0][0].shape[1]):
                n_att = len(attention[i][j - 1][:, head])
                n_supps = len(supps[i][j - 1])
                ax2.fill_between(
                    np.linspace(start=0, stop=n_supps, num=n_att),
                    attention[i][j - 1][:, head],
                    alpha=0.25,
                    # label=f"head {head}",
                    label="accumulated attention",
                )
            if j == 1:
                ax_base = ax2
            else:
                ax2.sharey(ax_base)
            ax2.set_ylabel("attention scores", color="red")
            ax2.legend()
    return fig


def plot_quantile_forecast(
    query_past: List[np.ndarray],
    query_future: List[np.ndarray],
    pred: List[np.ndarray],
    quantiles: List[str],
) -> plt.Figure:
    """
    Plots the provided forecasts for each sample with confidence intervals
    using all provided quantiles.

    Parameters
    ----------
    query_past: List[np.ndarray] of length n_samples, each array has shape [context length]
        The past query which the model uses to make a prediction.
    query_future: List[np.ndarray] of length n_samples, each array has shape [prediction horizon length]
        The ground truth for the forecast.
    pred: List[np.ndarray], of length n_samples, each array has shape  [prediction horizon length, n_quantiles]
        The prediction of the model.
    quantiles: List[float], shape[n_quantiles,]

    Returns
    -------
    plt.figure.Figure
        A plot containing one subplot for each sample. Each subplot displays the ground truth, predicted confidence
        intervals and the predicted median.
    """
    n_samples = len(query_past)
    fig, subs = plt.subplots(n_samples, figsize=(10, 4 * n_samples))
    if n_samples > 1:
        for i in range(n_samples):
            subs[i].set_title(f"sample {i}")
            _plot_quantile_forecast(
                subs[i], query_past[i], query_future[i], pred[i], quantiles
            )
    else:
        i = 0
        subs.set_title(f"sample {i}")
        _plot_quantile_forecast(
            subs, query_past[i], query_future[i], pred[i], quantiles
        )
    return fig


def _plot_quantile_forecast(
    sub: plt.Axes,
    query_past: np.ndarray,
    query_future: np.ndarray,
    pred: np.ndarray,
    quantiles: List[str],
):
    """
    Plots the provided forecast with confidence intervals using all provided
    quantiles.

    Parameters
    ----------
    sub: plt.Axis
        The axis object to plot on.
    query_past: np.ndarray, shape[context length]
        The past query which the model uses to make a prediction.
    query_future: np.ndarray, shape[prediction horizon length]
        The ground truth for the forecast.
    pred: np.ndarray, shape[prediction horizon length, n_quantiles]
        The prediction of the model.
    quantiles: List[str], shape[n_quantiles,]
    """
    assert len(pred) == len(
        query_future
    ), f"len pred is {len(pred)}, len query_future is {len(query_future)}"
    query = np.concatenate([query_past, query_future])
    sub.plot(query, label="gt")
    sub.axvline(len(query_past) - 1, color="r")  # end of train dataset
    pred_time = np.arange(len(query_past), len(query_past) + len(pred))
    cmap = plt.get_cmap("Oranges")
    # Plot the shapes for the confidence intervals, starting with the outermost
    num_intervals = pred.shape[-1] // 2
    for i in range(num_intervals):
        lower = pred[..., i]
        upper = pred[..., -(i + 1)]
        ci = 100 - 2 * int(float(quantiles[i]) * 100)
        cmap_index = int((cmap.N / (num_intervals + 1)) * (i + 1))
        if len(pred) > 1:
            sub.fill_between(
                pred_time,
                lower,
                upper,
                color=cmap(cmap_index),
                label=f"{ci}% CI",
            )
        else:
            sub.fill_between(
                np.append(pred_time, pred_time[0] + 0.5),
                np.repeat(lower, 2),
                np.repeat(upper, 2),
                color=cmap(cmap_index),
                label=f"{ci}% CI",
            )
    if len(pred) > 1:
        sub.plot(
            pred_time,
            pred[..., num_intervals],
            color="black",
            label="pred median",
        )
    else:
        sub.plot(
            np.append(pred_time, pred_time[0] + 0.5),
            np.repeat(pred[..., num_intervals], 2),
            color="black",
            label="pred median",
        )
    sub.legend()


def plot_point_forecast(sub, query_past, query_future, pred):
    query = np.concatenate([query_past, query_future])
    sub.plot(query, label="gt")
    sub.axvline(len(query_past) - 1, color="r")  # end of train dataset
    pred_time = np.arange(len(query_past), len(query_past) + len(pred))
    sub.plot(pred_time, pred, label="pred")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions
from torch.distributions import Normal
from abc import ABC
import numpy as np
import warnings

warnings.filterwarnings("ignore")

torch.autograd.set_detect_anomaly(False)
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

import os, pickle

hidden_dim = 100
sparsity = 100
sampling_size = 20


def save_object(obj, filename):
    with open(filename, "wb") as output:
        pickle.dump(obj, output)


def load_object(filename):
    with open(filename, "rb") as output:
        return pickle.load(output)


def check(FOLDER):
    if not os.path.exists(FOLDER):
        os.makedirs(FOLDER)


def seed(np_seed=11041987, torch_seed=20051987):
    np.random.seed(np_seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.manual_seed(torch_seed)


class SparseNet(
    nn.Module, ABC
):  # accept x, return w for which E[|| w ||_0] <= m -- our sparse layer
    def __init__(
        self,
        context_length,
        target_dim,
        target_item,
        hidden_dim,
        m,
        max_norm,
        norm=True,
    ):
        super(SparseNet, self).__init__()
        self.input_dim = context_length * target_dim
        self.context_length = context_length
        self.target_dim = target_dim
        self.hidden_dim = hidden_dim
        self.m = m
        self.norm = norm
        self.gamma = nn.Parameter(
            -1.0 * torch.ones((1, self.target_dim)), requires_grad=True
        ).to(device)
        self.fc1 = nn.Linear(self.input_dim, hidden_dim).to(device)
        self.fc21 = nn.Linear(hidden_dim, self.input_dim).to(device)
        self.fc22 = nn.Linear(hidden_dim, self.input_dim).to(device)
        self.bn = nn.BatchNorm1d(hidden_dim).to(device)
        self.dist = Normal(
            torch.FloatTensor([0.0]).to(device),
            torch.FloatTensor([1.0]).to(device),
        )
        self.target_item = target_item
        self.max_norm = max_norm
        # self.use_mean = use_mean

    def _r(self):  # output is of size 1 by input_dim
        return (
            (self.m / np.sqrt(self.target_dim))
            * torch.exp(0.5 * self.gamma)
            / torch.sqrt(torch.sum(torch.exp(self.gamma)))
        )

    def forward(self, x, n_sample=100):  # x is batch_size by input_dim
        x = x.view(x.shape[0], self.input_dim)
        x = self.fc1(x)
        x_reshape = x.view(-1, self.fc1.out_features)
        if (x_reshape.shape[0] > 1) and (self.norm is True):
            shape = x.shape
            x = x.view(-1, self.fc1.out_features)
            x = F.relu(self.bn(x)).view(shape)
        else:
            x = F.relu(x)
        mu, log_var = self.fc21(x), self.fc22(x)
        std = torch.exp(0.5 * log_var)
        eps = (
            torch.empty(n_sample, mu.shape[0], self.input_dim)
            .normal_(0, 1)
            .to(device)
        )
        w = mu + eps * std
        w = w.mean(0)
        u = torch.empty(mu.shape[0], self.target_dim).normal_(0, 1).to(device)
        r = self._r()
        r = r.repeat(mu.shape[0], 1)
        mask = (u <= self.dist.icdf(r)).to(device)  # batch x target_dim
        mask_target_item = torch.ones(mu.shape[0], self.target_dim).to(device)
        if self.target_item is not None:
            mask_target_item[:, self.target_item] = 0

        # renorm output
        output = w.view(-1, self.context_length, self.target_dim) * (
            mask.float() * mask_target_item
        ).reshape(-1, 1, self.target_dim).repeat(
            1, self.context_length, 1
        )  # batch x context x target dim
        output.clamp_(min=-self.max_norm, max=self.max_norm)
        return output.view(
            -1, self.context_length, self.target_dim
        )  # output size = batch_size * input_dim



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import os
from pathlib import Path
import numpy as np
import pandas as pd

from src.gluonts.dataset.repository._util import metadata, save_to_file, to_dict

from meta.datasets.gluonts import GluonTSDataModule
from meta.datasets.registry import register_data_module


M4_PREDICTION_LENGTHS = {
    "Hourly": 48,
    "Daily": 14,
    "Weekly": 13,
    "Monthly": 18,
    "Quarterly": 8,
    "Yearly": 6,
}


@register_data_module
class M4DataModule(GluonTSDataModule):
    """
    A data module which provides a frequency-category split of the M4 dataset
    as a standalone dataset.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.freq, self.category = [
            s.capitalize() for s in self.dataset_name.split("_")[1:]
        ]

    def _materialize(self, directory: Path) -> None:
        generate_m4_dataset(
            dataset_path=directory / self.dataset_name,
            m4_freq=self.freq,
            pandas_freq=self.freq[0],
            prediction_length=M4_PREDICTION_LENGTHS[self.freq],
            category=self.category,
        )

    @classmethod
    def name(cls) -> str:
        return "dm_m4"


def generate_m4_dataset(
    dataset_path: Path,
    m4_freq: str,
    pandas_freq: str,
    prediction_length: int,
    category: str,
):
    m4_dataset_url = (
        "https://github.com/M4Competition/M4-methods/raw/master/Dataset"
    )
    meta_df = pd.read_csv(f"{m4_dataset_url}/M4-info.csv", index_col=0)
    meta_df = meta_df.loc[meta_df["SP"] == m4_freq]

    train_df = pd.read_csv(
        f"{m4_dataset_url}/Train/{m4_freq}-train.csv", index_col=0
    )
    train_df = train_df.loc[meta_df["category"] == category]

    test_df = pd.read_csv(
        f"{m4_dataset_url}/Test/{m4_freq}-test.csv", index_col=0
    )
    test_df = test_df.loc[meta_df["category"] == category]

    meta_df = meta_df.loc[meta_df["category"] == category]

    os.makedirs(dataset_path, exist_ok=True)

    with open(dataset_path / "metadata.json", "w") as f:
        f.write(
            json.dumps(
                metadata(
                    cardinality=len(train_df),
                    freq=pandas_freq,
                    prediction_length=prediction_length,
                )
            )
        )

    train_file = dataset_path / "train" / "data.json"
    test_file = dataset_path / "test" / "data.json"

    train_target_values = [ts[~np.isnan(ts)] for ts in train_df.values]

    test_target_values = [
        np.hstack([train_ts, test_ts])
        for train_ts, test_ts in zip(train_target_values, test_df.values)
    ]

    if m4_freq == "Yearly":
        # some time series have more than 300 years which can not be
        # represented in pandas, this is probably due to a misclassification
        # of those time series as Yearly. We use only those time series with
        # fewer than upper limit items for this reason.
        start_dates = list(meta_df.StartingDate)
        upper_limit = pd.Timestamp.today().year - min(
            [pd.Timestamp(s).year for s in start_dates]
        )
        filter_long = [len(ts) <= upper_limit for ts in test_target_values]
        train_target_values = [
            ts for ts, b in zip(train_target_values, filter_long) if b
        ]
        test_target_values = [
            ts for ts, b in zip(test_target_values, filter_long) if b
        ]
        meta_df = meta_df.loc[filter_long]

    start_dates = list(meta_df.StartingDate)
    start_dates = [
        (
            sd
            if pd.Timestamp(sd) <= pd.Timestamp("2022")
            else str(pd.Timestamp("1900"))
        )
        for sd in start_dates
    ]

    save_to_file(
        train_file,
        [
            to_dict(
                target_values=target,
                start=start_date,
                cat=[cat],
                item_id=cat,
            )
            for cat, (target, start_date) in enumerate(
                zip(train_target_values, start_dates)
            )
        ],
    )

    save_to_file(
        test_file,
        [
            to_dict(
                target_values=target,
                start=start_date,
                cat=[cat],
                item_id=cat,
            )
            for cat, (target, start_date) in enumerate(
                zip(test_target_values, start_dates)
            )
        ],
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
import logging
from typing import Callable, Optional, List, Tuple
import pandas as pd
from autogluon.tabular import TabularPredictor as AutogluonTabularPredictor

from src.gluonts.core.component import validated
from src.gluonts.dataset.common import Dataset
from src.gluonts.dataset.util import to_pandas
from src.gluonts.model.estimator import Estimator
from src.gluonts.time_feature import (
    TimeFeature,
    get_lags_for_frequency,
    time_features_from_frequency_str,
)

from .predictor import (
    TabularPredictor,
    mean_abs_scaling,
    get_features_dataframe,
)

logger = logging.getLogger(__name__)


class TabularEstimator(Estimator):
    """
    An estimator that trains an Autogluon Tabular model for time series
    forecasting.

    Additional keyword arguments to the constructor, other than the ones documented
    below, will be passed on to Autogluon Tabular's ``fit`` method used for training
    the model.

    Parameters
    ----------
    freq
        Frequency of the data to handle
    prediction_length
        Prediction length
    lag_indices
        List of indices of the lagged observations to use as features. If
        None, this will be set automatically based on the frequency.
    time_features
        List of time features to be used. If None, this will be set automatically
        based on the frequency.
    scaling
        Function to be used to scale time series. This should take a pd.Series object
        as input, and return a scaled pd.Series and the scale (float). By default,
        this divides a series by the mean of its absolute value.
    batch_size
        Batch size of the resulting predictor; this is just used at prediction
        time, and does not affect training in any way.
    disable_auto_regression
        Whether to forcefully disable auto-regression in the model. If ``True``,
        this will remove any lag index which is smaller than ``prediction_length``.
        This will make predictions more efficient, but may impact their accuracy.
    quantiles_to_predict
        Whether to forecast in quantile way. If assigned with quantile values,
        this will train model using quantile prediction model. If None, then the model
        will be trained in a regular way.
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        lag_indices: Optional[List[int]] = None,
        time_features: Optional[List[TimeFeature]] = None,
        scaling: Callable[
            [pd.Series], Tuple[pd.Series, float]
        ] = mean_abs_scaling,
        batch_size: Optional[int] = 32,
        disable_auto_regression: bool = False,
        last_k_for_val: Optional[int] = None,
        quantiles_to_predict: Optional[List[float]] = None,
        eval_metric: str = "mean_absolute_error",
        **kwargs,
    ) -> None:
        super().__init__()

        self.prediction_length = prediction_length
        self.lag_indices = (
            lag_indices
            if lag_indices is not None
            else get_lags_for_frequency(freq)
        )
        self.time_features = (
            time_features
            if time_features is not None
            else time_features_from_frequency_str(freq)
        )
        self.batch_size = batch_size
        self.disable_auto_regression = disable_auto_regression
        self.scaling = scaling
        self.last_k_for_val = last_k_for_val
        self.eval_metric = eval_metric
        self.quantiles_to_predict = quantiles_to_predict

        if self.disable_auto_regression:
            self.lag_indices = [
                lag_idx
                for lag_idx in self.lag_indices
                if lag_idx >= self.prediction_length
            ]

        default_kwargs = {
            "time_limit": 60,
            # "excluded_model_types": ["KNN", "XT", "RF"],
            "presets": [
                "high_quality_fast_inference_only_refit",
                "optimize_for_deployment",
            ],
            "auto_stack": True,
        }
        self.kwargs = {**default_kwargs, **kwargs}

    def train(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
    ) -> TabularPredictor:
        kwargs_override = {}

        dfs = [
            get_features_dataframe(
                series=self.scaling(to_pandas(entry))[0],
                time_features=self.time_features,
                lag_indices=self.lag_indices,
            )
            for entry in training_data
        ]

        if validation_data is not None or self.last_k_for_val is not None:
            kwargs_override["auto_stack"] = False
            logger.warning(
                "Auto Stacking is turned off as validation dataset is provided"
                " before input into Tabular Predictor."
            )

        if validation_data is not None:
            logger.log(20, "Validation dataset is directly provided.")
            validation_dfs = [
                get_features_dataframe(
                    series=self.scaling(to_pandas(entry))[0],
                    time_features=self.time_features,
                    lag_indices=self.lag_indices,
                )
                for entry in validation_data
            ]
            train_df = pd.concat(dfs)
            val_df = pd.concat(validation_dfs)
        elif self.last_k_for_val is not None:
            logger.log(
                20,
                "last_k_for_val is provided, choosing last"
                f" {self.last_k_for_val} of each time series as validation"
                " set.",
            )
            train_dfs = [
                tmp_df.iloc[: -self.last_k_for_val, :] for tmp_df in dfs
            ]
            validation_dfs = [
                tmp_df.iloc[-self.last_k_for_val :, :] for tmp_df in dfs
            ]
            train_df = pd.concat(train_dfs)
            val_df = pd.concat(validation_dfs)
        else:
            logger.log(
                20,
                "No validation dataset is provided, will let TabularPredictor"
                " do the splitting automatically,Note that this might break"
                " the time order of time series data.",
            )
            train_df = pd.concat(dfs)
            val_df = None

        if self.quantiles_to_predict is not None:
            ag_model = AutogluonTabularPredictor(
                label="target",
                problem_type="quantile",
                quantile_levels=self.quantiles_to_predict,
            ).fit(
                train_df,
                tuning_data=val_df,
                **{**self.kwargs, **kwargs_override},
            )
        else:
            ag_model = AutogluonTabularPredictor(
                label="target",
                problem_type="regression",
                eval_metric=self.eval_metric,
            ).fit(
                train_df,
                tuning_data=val_df,
                **{**self.kwargs, **kwargs_override},
            )

        return TabularPredictor(
            ag_model=ag_model,
            prediction_length=self.prediction_length,
            time_features=self.time_features,
            lag_indices=self.lag_indices,
            scaling=self.scaling,
            batch_size=self.batch_size,
            quantiles_to_predict=self.quantiles_to_predict,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import torch
from torch.distributions import (
    Distribution,
    TransformedDistribution,
    AffineTransform,
)


class ImplicitQuantile(Distribution):
    arg_constraints = {}

    def __init__(
        self,
        implicit_quantile_function,
        taus,
        nn_output,
        predicted_quantiles,
        validate_args=None,
    ):
        self.predicted_quantiles = predicted_quantiles[0]
        self.taus = taus
        self.quantile_function = implicit_quantile_function
        self.input_data = nn_output

        super(ImplicitQuantile, self).__init__(
            batch_shape=self.predicted_quantiles.shape,
            validate_args=validate_args,
        )

    @torch.no_grad()
    def sample(self, sample_shape=torch.Size()):
        """
        See arXiv: 1806.06923 Once the model has learned how to predict a given
        quantile tau, one can sample from the distribution of the target, by
        sampling tau values.
        """
        if len(sample_shape) == 0:
            num_parallel_samples = 1
        else:
            num_parallel_samples = sample_shape[0]
        input_data = torch.repeat_interleave(
            self.input_data, repeats=num_parallel_samples, dim=0
        )
        batch_size = input_data.shape[0]
        forecast_length = input_data.shape[1]
        device = input_data.device

        taus = torch.rand(size=(batch_size, forecast_length), device=device)
        samples = self.quantile_function(input_data, taus)
        if len(sample_shape) == 0:
            return samples
        else:
            return samples.reshape((num_parallel_samples, -1, forecast_length))

    def log_prob(self, value):
        # Assumes same distribution for all steps in the future, conditionally on the input data
        return -self.quantile_loss(self.predicted_quantiles, value, self.taus)

    @staticmethod
    def quantile_loss(quantile_forecast, target, tau):
        return torch.abs(
            (quantile_forecast - target)
            * ((target <= quantile_forecast).float() - tau)
        )


class TransformedImplicitQuantile(TransformedDistribution):
    def __init__(self, base_distribution, transforms):
        super().__init__(base_distribution, transforms)

    def log_prob(self, x):
        scale = 1.0
        for transform in reversed(self.transforms):
            assert isinstance(
                transform, AffineTransform
            ), "Not an AffineTransform"
            x = transform.inv(x)
            scale *= transform.scale
        p = self.base_dist.log_prob(x)
        return p * scale



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass
from typing import cast, Generic, List, TypeVar
import pandas as pd
from tsbench.config import Config, EnsembleConfig, ModelConfig
from tsbench.evaluations.metrics import Performance

T = TypeVar("T", ModelConfig, EnsembleConfig)


@dataclass
class Evaluations(Generic[T]):
    """
    Generic base for evaluations of models.
    """

    configurations: List[Config[T]]
    performances: List[Performance]

    def dataframe(self, std: bool = True) -> pd.DataFrame:
        """
        Returns a dataframe which contains the performance metrics as columns
        and the configurations as multi-index.

        Args:
            std: Whether to include the standard deviation of performance metrics in the dataframe.
        """
        # Should implement this for ensembles as well
        index_df = Config.to_dataframe(
            cast(List[Config[ModelConfig]], self.configurations)
        )
        # Reorder columns
        column_order = ["dataset"] + [
            c for c in index_df.columns.tolist() if c != "dataset"
        ]
        index = pd.MultiIndex.from_frame(index_df[column_order])
        df = Performance.to_dataframe(self.performances, std=std)
        df.index = index
        return df.sort_index()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Callable, Any
import torch
from torchmetrics import Metric


class QuantileLoss(Metric):
    """
    Computes the quantile loss.

    Parameters
    ----------
    quantiles: The quantiles.
    """

    def __init__(
        self,
        quantiles: List[str],
        compute_on_step: bool = True,
        dist_sync_on_step: bool = False,
        process_group: Optional[Any] = None,
        dist_sync_fn: Callable = None,
    ):
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )
        self.register_buffer(
            "quantiles", torch.as_tensor([float(q) for q in quantiles])
        )
        self.add_state(
            "quantile_sum", default=torch.as_tensor(0.0), dist_reduce_fx="sum"
        )
        self.add_state(
            "num_observations",
            default=torch.as_tensor(0.0),
            dist_reduce_fx="sum",
        )

    def update(
        self, y_pred: torch.Tensor, y_true: torch.Tensor, mask: torch.Tensor
    ) -> None:
        # crop y_pred to shape of y_true since the model always predicts the max prediction length
        # max prediction length might be longer than max length in the particular batch
        y_pred = y_pred[:, : y_true.size()[1], ...]
        y_true = y_true.unsqueeze(-1)
        losses = torch.sum(
            torch.abs(
                (y_pred - y_true)
                * ((y_true <= y_pred).type(torch.uint8) - self.quantiles)
            ),
            axis=-1,
        )
        # mask out all elements that correspond to padding in y_true
        losses = torch.mul(losses, mask)
        self.quantile_sum += losses.sum()
        self.num_observations += mask.sum() * len(self.quantiles)

    def compute(self) -> torch.Tensor:
        return self.quantile_sum / self.num_observations



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractmethod
import torch
from meta.models.EcDc.components.tcn import CausalCNNEncoder
from torch import nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from meta.data.batch import SeriesBatch


class SupportSetEncoder(nn.Module, ABC):
    """
    Base class for support set encoders.
    """

    @abstractmethod
    def forward(self, supps: SeriesBatch) -> SeriesBatch:
        """
        Encodes each time point of each support set time series with a vector
        of fixed size.

        Parameters
        ----------
        supps:
            A SeriesBatch containing sequences of size [batch, sequence length, n_features].

        Returns
        -------
        SeriesBatch:
            The transformed batch containing sequences with the encoded support set series
            of size [batch, sequence length, encoding size].
        """


class LSTMSupportSetEncoder(SupportSetEncoder):
    """
    Encodes each time step in the support set times series via the hidden
    states of a LSTM.

    Parameters
    ----------
    input_size: The number of features of the time series.
    hidden_size: The size of hidden states of LSTM, corresponds to encoding size of the queries,
        i.e. the encoder returns a tensor with shape [batch, sequence length, D * hidden_size].
    num_layers: number of layers of the LSTM.
    bidirectional: If a bidirectional LSTM is used. If true: D = 2, else D = 1.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int,
        bidirectional: bool,
    ):
        super().__init__()
        self.encoder = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            num_layers=num_layers,
            bidirectional=bidirectional,
            batch_first=True,
        )

    def forward(self, supps: SeriesBatch) -> torch.Tensor:
        supps_packed = pack_padded_sequence(
            supps.sequences,
            supps.lengths.cpu(),
            batch_first=True,
            enforce_sorted=False,
        )
        output_packed, _ = self.encoder(supps_packed)
        output_padded, output_lengths = pad_packed_sequence(
            output_packed, batch_first=True
        )
        return SeriesBatch(output_padded, output_lengths, supps.split_sections)


class CNNSupportSetEncoder(SupportSetEncoder):
    """
    Encodes the time series in the support set via a CNN.

    Parameters
    ----------
    input_size: The number of features of the time series.
    out_channels: The number of output channels of the CNN, corresponds to encoding size of the queries,
        i.e. the encoder returns a tensor with shape [batch, sequence length, D * hidden_size].
    """

    def __init__(self, input_size: int, out_channels: int):
        super().__init__()
        self.conv1 = nn.Conv1d(
            in_channels=input_size,
            out_channels=out_channels,
            kernel_size=5,
            padding="valid",
        )

        self.conv2 = nn.Conv1d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=5,
            padding="same",
        )
        self.conv3 = nn.Conv1d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            padding="same",
        )
        self.conv4 = nn.Conv1d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            padding="same",
        )

    def forward(self, supps: SeriesBatch) -> torch.Tensor:
        # TODO: permute should be handled as here:
        # https://pytorch.org/tutorials/intermediate/memory_format_tutorial.html
        # https://discuss.pytorch.org/t/dataset-with-last-dimension-as-channel/100639/6
        pool = F.avg_pool1d
        channels_first = supps.sequences.permute(0, 2, 1)
        x = pool(F.relu(self.conv1(channels_first)), 2)
        x = pool(F.relu(self.conv2(x)), 2)
        x = pool(F.relu(self.conv3(x)), 2)
        x = pool(F.relu(self.conv4(x)), 2)

        # TODO: padding of the batches is not takes into account!
        return SeriesBatch(
            x.permute(0, 2, 1), supps.lengths, supps.split_sections
        )


class TcnSupportSetEncoder(SupportSetEncoder):
    """
    Encodes each time point of the support set time series via WaveNet.

    Parameters
    ----------
    num_channels: The number of filters in each CNN layer.
    num_layers: The number of causal convolution blocks
    kernel_size: The kernel size of the filters in the causal convolution blocks
    """

    def __init__(
        self,
        encoder: nn.Module = None,
        num_channels: int = 64,
        num_layers: int = 5,
        kernel_size: int = 2,
    ):
        super().__init__()
        if encoder:
            self.encoder = encoder
        else:
            self.encoder = CausalCNNEncoder(
                in_channels=1,
                channels=num_channels,
                depth=num_layers,
                out_channels=num_channels,
                kernel_size=kernel_size,
            )

    def forward(self, supps: SeriesBatch) -> torch.Tensor:
        output_padded = self.encoder(supps.sequences.transpose(1, 2))
        return SeriesBatch(
            output_padded.transpose(1, 2), supps.lengths, supps.split_sections
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from collections import defaultdict
from typing import Dict, List
from tsbench.config import ModelConfig


def union_dicts(
    dicts: List[Dict[str, ModelConfig]]
) -> Dict[str, List[ModelConfig]]:
    """
    Merges the dicts by aggregating model configurations with the same key into
    a list.
    """
    result = defaultdict(list)
    for item in dicts:
        for k, v in item.items():
            result[k].append(v)
    return result



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import Optional, List, Tuple, Union, Iterator
from collections import defaultdict
import warnings
import math

import torch as pt
from torch import Tensor, LongTensor, BoolTensor
from torch import nn
from torch.nn import Parameter, functional as F

from .modules import AttentionEstimator, AdversarialEstimator
from ..network import AttentionKernel, AttentionBlock, AdversarialBlock


class DomAdaptEstimator(nn.Module):
    def __init__(
        self,
        src_module: AttentionEstimator,
        tgt_module: AttentionEstimator,
        balance_loss: bool = True,
        forecast_target: bool = True,
    ) -> None:
        super(DomAdaptEstimator, self).__init__()
        self.src = src_module
        self.tgt = tgt_module
        self.balance_loss = balance_loss
        self.forecast_target = forecast_target

    def forward(
        self,
        src_data: Tensor,
        tgt_data: Tensor,
        src_feats: Optional[Tensor] = None,
        tgt_feats: Optional[Tensor] = None,
        src_nan_mask: Optional[BoolTensor] = None,
        tgt_nan_mask: Optional[BoolTensor] = None,
        src_length: Optional[LongTensor] = None,
        tgt_length: Optional[LongTensor] = None,
    ) -> Tensor:
        src_loss = self.src(src_data, src_feats, src_nan_mask, src_length)
        tgt_loss = self.tgt(tgt_data, tgt_feats, tgt_nan_mask, tgt_length)
        if not self.forecast_target:
            tgt_loss = tgt_loss - self.tgt.tradeoff * self.tgt.fc_loss
        # after rescaling, balance two losses
        if self.balance_loss:
            src_scale = self.src._normalizer._buffers["scale"]
            tgt_scale = self.tgt._normalizer._buffers["scale"]
            src_scale = src_scale.view(src_scale.size(0), -1)
            tgt_scale = tgt_scale.view(tgt_scale.size(0), -1)
            weight = pt.mean(tgt_scale / src_scale, dim=1)
            src_loss = src_loss * weight
        loss = src_loss + tgt_loss
        return loss


class AdversarialDomAdaptEstimator(DomAdaptEstimator):
    def __init__(
        self,
        src_module: AdversarialEstimator,
        tgt_module: AdversarialEstimator,
        balance_loss: bool = True,
        forecast_target: bool = True,
        disc_lambda: float = 1.0,
    ) -> None:
        super(AdversarialDomAdaptEstimator, self).__init__(
            src_module,
            tgt_module,
            balance_loss,
            forecast_target,
        )
        self.disc_lambda = disc_lambda
        self._generative = True

    def generative(self):
        self._generative = True
        self.src.generative()
        self.tgt.generative()

    def discriminative(self):
        self._generative = False
        self.src.discriminative()
        self.tgt.discriminative()

    def forward(
        self,
        src_data: Tensor,
        tgt_data: Tensor,
        src_feats: Optional[Tensor] = None,
        tgt_feats: Optional[Tensor] = None,
        src_nan_mask: Optional[BoolTensor] = None,
        tgt_nan_mask: Optional[BoolTensor] = None,
        src_length: Optional[LongTensor] = None,
        tgt_length: Optional[LongTensor] = None,
    ) -> Tensor:
        gen_loss = super(AdversarialDomAdaptEstimator, self).forward(
            src_data,
            tgt_data,
            src_feats,
            tgt_feats,
            src_nan_mask,
            tgt_nan_mask,
            src_length,
            tgt_length,
        )

        batch_size = gen_loss.size(0)
        src_prob_domain = self.src.prob_domain.view(batch_size, -1)
        tgt_prob_domain = self.tgt.prob_domain.view(batch_size, -1)
        adv_type = "grad_rev"
        # adv_type = 'label_inv'

        # disc_loss = -log p(is_src | shape_src) - log(1 - p(is_src|shape_tgt))
        disc_loss = -src_prob_domain.mean(dim=1) - pt.log(
            1.0 - tgt_prob_domain.exp() + 1e-10
        ).mean(dim=1)
        if self._generative:
            if adv_type == "grad_rev":
                # DANN: gradient reversal operation
                # Generative model maximizes discriminative loss
                loss = gen_loss - self.disc_lambda * disc_loss
            elif adv_type == "label_inv":
                # ADDA: GAN-style minimax game
                # Invert the label of target data to make pseudo source samples
                # Do not update for real source samples
                # invert_tgt_loss = -log p(is_src | shape_tgt)
                invert_tgt_loss = -tgt_prob_domain.mean(dim=1)
                loss = gen_loss - self.disc_lambda * invert_tgt_loss
        else:
            loss = disc_loss

        return loss



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset

from src.gluonts.core.component import validated
from src.gluonts.time_feature import TimeFeature, norm_freq_str


class FourierDateFeatures(TimeFeature):
    @validated()
    def __init__(self, freq: str) -> None:
        super().__init__()
        # reocurring freq
        freqs = [
            "month",
            "day",
            "hour",
            "minute",
            "weekofyear",
            "weekday",
            "dayofweek",
            "dayofyear",
            "daysinmonth",
        ]

        assert freq in freqs
        self.freq = freq

    def __call__(self, index: pd.DatetimeIndex) -> np.ndarray:
        values = getattr(index, self.freq)
        num_values = max(values) + 1
        steps = [x * 2.0 * np.pi / num_values for x in values]
        return np.vstack([np.cos(steps), np.sin(steps)])


def fourier_time_features_from_frequency(freq_str: str) -> List[TimeFeature]:
    offset = to_offset(freq_str)
    granularity = norm_freq_str(offset.name)

    features = {
        "M": ["weekofyear"],
        "ME": ["weekofyear"],
        "W": ["daysinmonth", "weekofyear"],
        "D": ["dayofweek"],
        "B": ["dayofweek", "dayofyear"],
        "H": ["hour", "dayofweek"],
        "h": ["hour", "dayofweek"],
        "min": ["minute", "hour", "dayofweek"],
        "T": ["minute", "hour", "dayofweek"],
    }

    assert granularity in features, f"freq {granularity} not supported"

    feature_classes: List[TimeFeature] = [
        FourierDateFeatures(freq=freq) for freq in features[granularity]
    ]
    return feature_classes



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx

from src.gluonts.mx import Tensor


class GNN(mx.gluon.HybridBlock):
    def __init__(
        self,
        units: int,
        num_layers: int,
        adj_matrix: Tensor,
        use_mlp: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.units = units
        self.num_layers = num_layers
        self.adj_matrix = adj_matrix
        self.use_mlp = use_mlp

        if self.use_mlp:
            with self.name_scope():
                self.gnn_layer = mx.gluon.nn.Dense(
                    units=self.units, flatten=False
                )

    def hybrid_forward(self, F, x, *args, **kwargs):
        # Do message passing for `num_layers` times with learnable weights.
        for _ in range(self.num_layers):
            if self.use_mlp:
                x = x + self.gnn_layer(x)
                x = F.dot(x.swapaxes(-1, -2), self.adj_matrix).swapaxes(-1, -2)
                x = F.relu(x)
            else:
                x = F.dot(x.swapaxes(-1, -2), self.adj_matrix).swapaxes(-1, -2)

        return x



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import subprocess
from pathlib import Path
from typing import Any


def run_sacred_script(script: str, **kwargs: Any) -> None:
    """
    Runs the Sacred script with the provided name in a subprocess, passing the
    provided parameters.

    Args:
        script: The name of the script in `cli/sacred`.
        kwargs: Parameters to pass to the script.
    """
    subprocess.call(
        [
            "poetry",
            "run",
            "python",
            "-W",
            "ignore",
            Path(os.path.realpath(__file__)).parent.parent
            / "analysis"
            / "scripts"
            / script,
            "-m",
            "localhost:27017:sacred",
            "with",
        ]
        + [f"{k}={_sacred_value(v)}" for k, v in kwargs.items()],
        cwd=Path(os.path.realpath(__file__)).parent.parent.parent.parent,
    )


def _sacred_value(value: Any) -> str:
    return f'"{value}"' if isinstance(value, str) else f"{value}"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass, asdict
from typing import Optional, Literal, Tuple
from pathlib import Path
from tqdm.auto import tqdm
import json
import numpy as np
import random
import pytorch_lightning as pl
from src.gluonts.dataset.field_names import FieldName  # type: ignore
from src.gluonts.dataset.repository.datasets import get_dataset, materialize_dataset  # type: ignore
from torch.utils.data import DataLoader
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF
import warnings

warnings.filterwarnings("ignore")

from meta.data.batch import TripletBatch
from meta.datasets.splits import DatasetSplits
from meta.data.sampling import TripletDataset
from meta.datasets.registry import register_data_module


@dataclass
class MetaData:
    """
    Meta data for artificial datasets.
    """

    context_length_multiple: int
    support_set_size: int
    prediction_length: int
    freq: str
    support_length_multiple: int = 4
    num_queries: int = 1

    @classmethod
    def parse_file(cls, file: Path):
        with open(file) as json_file:
            return MetaData(**json.load(json_file))

    def save(self, file: Path):
        with open(file, "w") as fp:
            json.dump(asdict(self), fp)


@register_data_module
class ArtificialDataModule(pl.LightningDataModule):
    """
    A data module which provides an artificial dataset.
    """

    def __init__(
        self,
        dataset_name: str,
        context_length_multiple: int,
        support_length_multiple: int,
        support_set_size: int,
        num_queries: int = 1,
        data_dir: Path = Path.home() / ".mxnet" / "gluont-ts",
        prediction_length: Optional[int] = None,
        test_set_size: int = 200,
        val_set_size: int = 200,
        train_set_size: int = 10000,
        standardize: bool = False,
        batch_size_train: int = 128,
        batch_size_val: int = 1000,
        batch_size_test: int = 1000,
        num_workers: int = 0,
        seed: Optional[int] = None,
    ):
        super().__init__()
        self.dataset_name = dataset_name
        self.data_dir = data_dir
        self.context_length_multiple = context_length_multiple
        self.support_length_multiple = support_length_multiple
        self._prediction_length = prediction_length
        self.test_set_size = test_set_size
        self.val_set_size = val_set_size
        self.train_set_size = train_set_size
        self.support_set_size = support_set_size
        self.num_queries = num_queries
        self.standardize = standardize
        self.splits: DatasetSplits
        self.batch_size_train = batch_size_train
        self.batch_size_val = batch_size_val
        self.batch_size_test = batch_size_test
        self.num_workers = num_workers
        self.seed = seed

    @property
    def context_length(self) -> int:
        return self.context_length_multiple * self.meta.prediction_length

    @property
    def support_length(self) -> int:
        return self.support_length_multiple * self.prediction_length

    @property
    def prediction_length(self) -> int:
        return self._prediction_length or self.meta.prediction_length

    @property
    def root(self) -> Path:
        """
        Returns the directory where all the data pertaining to this dataset is
        stored.
        """
        return self.data_dir / "artificial" / self.dataset_name

    @property
    def meta(self) -> MetaData:
        """
        Returns the dataset's metadata.
        """
        return (
            MetaData.parse_file(self.root / "metadata.json")
            if self.root.exists()
            else None
        )

    def setup(self, stage: Optional[str] = None) -> None:
        self.generate()
        self.splits = DatasetSplits(
            self.meta,
            self.root,
            self.dataset_name,
            self.prediction_length,
            self.standardize,
        )
        self.splits.train().prepare()
        self.splits.val().prepare()
        self.splits.test().prepare()

    def train_dataloader(self) -> DataLoader[TripletBatch]:
        split = self.splits.train()
        return DataLoader(
            TripletDataset(
                queries=split.data(), support_sets=split.support_set()
            ),
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_train,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def val_dataloader(self) -> DataLoader[TripletBatch]:
        split = self.splits.val()
        return DataLoader(
            TripletDataset(
                queries=split.data(), support_sets=split.support_set()
            ),
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_val,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def test_dataloader(self) -> DataLoader[TripletBatch]:
        split = self.splits.train()
        return DataLoader(
            TripletDataset(
                queries=split.data(), support_sets=split.support_set()
            ),
            collate_fn=TripletBatch.collate,  # type: ignore
            batch_size=self.batch_size_test,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def generate(self) -> None:
        meta = MetaData(
            context_length_multiple=self.context_length_multiple,
            support_length_multiple=self.support_length_multiple,
            support_set_size=self.support_set_size,
            prediction_length=self.prediction_length,
            num_queries=self.num_queries,
            freq="1H",
        )
        if self.root.exists():
            # check if the meta data of the dataset fits the requirements
            if self.meta == meta:
                return
            else:
                raise ValueError(
                    "Meta data of artificial dataset not compatible with requirements"
                )

        self.root.mkdir(parents=True, exist_ok=True)
        meta.save(self.root / "metadata.json")
        for split, n_samples in zip(
            ["train", "val", "test"],
            [self.train_set_size, self.val_set_size, self.test_set_size],
        ):
            self.generate_split(split, n_samples)

    def generate_split(
        self, split: Literal["train", "val", "test"], n_samples: int
    ) -> None:
        queries, support_sets = zip(
            *[
                generate_artificial_tuplets(
                    dataset_name=self.dataset_name,
                    context_length=self.context_length,
                    support_length=self.support_length,
                    prediction_length=self.prediction_length,
                    support_set_size=self.support_set_size,
                    item_id=i,
                )
                for i in tqdm(
                    range(n_samples), desc="generating artificial data"
                )
            ]
        )
        _write_data_to_file(self.root / split / "data.json", queries)
        _write_data_to_file(
            self.root / split / ".support_set.json", support_sets
        )

    @classmethod
    def name(cls) -> str:
        return "dm_artificial"

    def evaluate_model(self, **kwargs):
        pass


def _write_data_to_file(file: Path, data: Tuple) -> None:
    file.parent.mkdir(parents=True, exist_ok=True)
    with file.open("w") as f:
        content = "\n".join([json.dumps(d) for d in data])
        f.write(content + "\n")


def generate_artificial_tuplets(
    dataset_name: str,
    context_length: int,
    support_length: int,
    prediction_length: int,
    support_set_size: int,
    item_id: int,
) -> Tuple:
    if dataset_name == "marker":
        noise_level = 0.01
        context = np.random.normal(0, noise_level, context_length)
        signal_type = random.choice([-1, 1])
        target = signal_type * np.arange(0, 1, 1 / prediction_length)
        query = np.concatenate((context, target))

        # support time series without the marker
        support_set = list(
            np.random.normal(0, noise_level, context_length)
            for _ in range(support_set_size - 1)
        )
        # build support time series which contains the marker
        marker = (
            (-1)
            * signal_type
            * np.concatenate(
                (np.arange(0, 1, 1 / 6), np.flip(np.arange(0, 1, 1 / 6)))
            )
        )
        marker_start = np.random.choice(
            np.arange(0, context_length - len(marker))
        )
        support_ts = np.random.normal(0, noise_level, context_length)
        support_ts[marker_start : marker_start + len(marker)] += marker
        # insert at random position in support set
        support_set.insert(np.random.choice(support_set_size), support_ts)

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s

    if dataset_name == "marker_v2":
        noise_level = 0.01
        context = np.random.normal(0, noise_level, context_length)
        signal_type = random.choice([-1, 1])
        target = signal_type * np.arange(0, 1, 1 / prediction_length)
        query = np.concatenate((context, target))

        # support time series without the marker
        support_set = list(
            np.random.normal(0, noise_level, context_length)
            for _ in range(support_set_size - 1)
        )
        # build support time series which contains the marker
        ramp = np.arange(0, 1, 1 / 6)
        if signal_type == -1:
            marker = np.concatenate((ramp, np.flip(ramp)))
        else:
            marker = np.concatenate(
                (ramp, np.array([1, 0.8, 1]), np.flip(ramp))
            )
        marker_start = np.random.choice(
            np.arange(0, context_length - len(marker))
        )
        support_ts = np.random.normal(0, noise_level, context_length)
        support_ts[marker_start : marker_start + len(marker)] += marker
        # insert at random position in support set
        support_set.insert(np.random.choice(support_set_size), support_ts)

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "marker_v3":
        noise_level = 0.01
        context = np.random.normal(0, noise_level, context_length)
        signal_type = random.choice([-1, 0, 1])
        target = signal_type * np.arange(0, 1, 1 / prediction_length)
        query = np.concatenate((context, target))

        # support time series without the marker
        support_set = list(
            np.random.normal(0, noise_level, context_length)
            for _ in range(support_set_size - 1)
        )
        # build support time series which contains the marker
        ramp = np.arange(0, 1, 1 / 6)
        if signal_type == -1:
            marker = np.concatenate((ramp, np.flip(ramp)))
        elif signal_type == 1:
            marker = np.concatenate(
                (ramp, np.array([1, 0.8, 1]), np.flip(ramp))
            )
        else:
            marker = np.concatenate(
                (ramp, np.array([1, 1, 1, 1, 1]), np.flip(ramp))
            )
        marker_start = np.random.choice(
            np.arange(0, context_length - len(marker))
        )
        support_ts = np.random.normal(0, noise_level, context_length)
        support_ts[marker_start : marker_start + len(marker)] += marker
        # insert at random position in support set
        support_set.insert(np.random.choice(support_set_size), support_ts)

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "marker_v4":
        noise_level = 0.01
        context = np.random.normal(0, noise_level, context_length)
        signal_type = random.choice([-1, 0, 1])
        target = signal_type * np.arange(0, 1, 1 / prediction_length)
        query = np.concatenate((context, target))

        # support time series without the marker
        support_set = list(
            np.random.normal(0, noise_level, context_length)
            for _ in range(support_set_size - 1)
        )
        # build support time series which contains the marker
        ramp = np.arange(0, 1, 1 / 3)
        if signal_type == -1:
            marker = np.concatenate((ramp, np.flip(ramp)))
        elif signal_type == 1:
            marker = np.concatenate(
                (ramp, np.array([1, 0.8, 1]), np.flip(ramp))
            )
        else:
            marker = np.concatenate((ramp, np.array([1, 1, 1]), np.flip(ramp)))
        marker_start = np.random.choice(
            np.arange(0, context_length - len(marker))
        )
        support_ts = np.random.normal(0, noise_level, context_length)
        support_ts[marker_start : marker_start + len(marker)] += marker
        # insert at random position in support set
        support_set.insert(np.random.choice(support_set_size), support_ts)

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "marker_v6":
        noise_level = 0.01
        context = np.random.normal(0, noise_level, context_length)
        signal_type = random.choice([-1, 0, 1])
        target = signal_type * np.arange(0, 1, 1 / prediction_length)
        query = np.concatenate((context, target))

        # support time series without the marker
        support_set = list(
            np.random.normal(0, noise_level, context_length)
            for _ in range(support_set_size - 1)
        )
        # build support time series which contains the marker
        markers = [4 * [0.2], [0, 0.2, 0.2, 0], [0, 0.15, 0.2, 0.25]]
        marker = markers[signal_type + 1]
        marker_start = np.random.choice(
            np.arange(0, context_length - len(marker))
        )
        support_ts = np.random.normal(0, noise_level, context_length)
        support_ts[marker_start : marker_start + len(marker)] = marker
        # insert at random position in support set
        support_set.insert(np.random.choice(support_set_size), support_ts)
        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "cheat_random":
        context = np.random.normal(0, 0.001, context_length)
        target = np.random.normal(0, 0.1, prediction_length)
        query = np.concatenate((context, target))
        support_set = list(
            np.random.normal(0, 0.001, context_length)
            for _ in range(support_set_size)
        )
        si = np.random.choice(support_set_size)
        support_set[si][-prediction_length:] = query[-prediction_length:]
        # insert at random position in support set
        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "cheat_random_gp":
        kernel = 0.5 * RBF(length_scale=5.0, length_scale_bounds=(1e-1, 10.0))
        gpr = GaussianProcessRegressor(kernel=kernel, random_state=item_id)
        query = gpr.sample_y(
            np.arange(prediction_length + context_length).reshape(-1, 1),
            1,
            random_state=item_id + np.random.randint(100000),
        ).squeeze()

        marker_length = prediction_length

        support_set = gpr.sample_y(
            np.arange(support_length).reshape(-1, 1),
            support_set_size,
            random_state=item_id + np.random.randint(100000),
        ).T
        signal_length = marker_length + prediction_length
        marker_start = np.random.choice(
            np.arange(0, support_length - signal_length)
        )
        si = np.random.choice(support_set_size)
        support_set[si][marker_start : marker_start + signal_length] = query[
            -signal_length:
        ]

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "cheat_random_gp_fitted":
        kernel = 0.5 * RBF(length_scale=5.0, length_scale_bounds=(1e-1, 10.0))
        gpr = GaussianProcessRegressor(kernel=kernel, random_state=item_id)
        query = gpr.sample_y(
            np.arange(prediction_length + context_length).reshape(-1, 1),
            1,
            random_state=item_id + np.random.randint(100000),
        ).squeeze()

        marker_length = prediction_length

        support_set = gpr.sample_y(
            np.arange(context_length).reshape(-1, 1),
            support_set_size,
            random_state=item_id + np.random.randint(100000),
        ).T
        signal_length = marker_length + prediction_length
        marker_start = np.random.choice(
            np.arange(0, context_length - signal_length)
        )
        si = np.random.choice(support_set_size)
        x = np.arange(marker_start, marker_start + signal_length)
        gpr.fit(x.reshape(-1, 1), query[-signal_length:].reshape(-1, 1))
        s_ts = gpr.sample_y(
            np.arange(context_length).reshape(-1, 1),
            1,
            random_state=item_id + np.random.randint(100000),
        )
        support_set[si] = s_ts.squeeze()

        # insert at random position in support set
        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "cheat_random_gp_mix_marker":
        signal_type = random.choice(["gp", "marker"])
        kernel = 0.5 * RBF(length_scale=5.0, length_scale_bounds=(1e-1, 10.0))
        gpr = GaussianProcessRegressor(kernel=kernel, random_state=item_id)
        query = gpr.sample_y(
            np.arange(prediction_length + context_length).reshape(-1, 1),
            1,
            random_state=item_id + np.random.randint(100000),
        ).squeeze()
        support_set = gpr.sample_y(
            np.arange(context_length).reshape(-1, 1),
            support_set_size,
            random_state=item_id + np.random.randint(100000),
        ).T
        if signal_type == "gp":
            marker_length = prediction_length
            signal_length = marker_length + prediction_length
            marker_start = np.random.choice(
                np.arange(0, context_length - signal_length)
            )
            si = np.random.choice(support_set_size)
            support_set[si][marker_start : marker_start + signal_length] = (
                query[-signal_length:]
            )
        else:
            signal = np.concatenate(
                (np.ones((4,)), query[-prediction_length:])
            )
            signal_length = prediction_length + 4
            marker_start = np.random.choice(
                np.arange(0, context_length - signal_length)
            )
            si = np.random.choice(support_set_size)

            support_set[si][
                marker_start : marker_start + signal_length
            ] = signal
        # insert at random position in support set
        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "cheat_random_gp_mix":
        signal_type = random.choices(
            ["gp", "marker"], weights=[0.5, 0.5], k=1
        )[0]
        kernel = 0.5 * RBF(length_scale=5.0, length_scale_bounds=(1e-1, 10.0))
        gpr = GaussianProcessRegressor(kernel=kernel, random_state=item_id)
        query = gpr.sample_y(
            np.arange(prediction_length + context_length).reshape(-1, 1),
            1,
            random_state=item_id + np.random.randint(100000),
        ).squeeze()
        support_set = gpr.sample_y(
            np.arange(context_length).reshape(-1, 1),
            support_set_size,
            random_state=item_id + np.random.randint(100000),
        ).T
        if signal_type == "gp":
            marker_length = prediction_length
            signal_length = marker_length + prediction_length
            marker_start = np.random.choice(
                np.arange(0, context_length - signal_length)
            )
            si = np.random.choice(support_set_size)
            support_set[si][marker_start : marker_start + signal_length] = (
                query[-signal_length:]
            )
        # else:
        #     signal = query[-prediction_length:]
        #     signal_length = prediction_length
        #     marker_start = np.random.choice(np.arange(0, context_length - signal_length))
        #     si = np.random.choice(support_set_size)

        #     support_set[si][marker_start : marker_start + signal_length] = signal
        # insert at random position in support set
        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s
    if dataset_name == "random_gp":
        kernel = 0.5 * RBF(length_scale=5.0, length_scale_bounds=(1e-1, 10.0))
        gpr = GaussianProcessRegressor(kernel=kernel, random_state=item_id)
        query = gpr.sample_y(
            np.arange(prediction_length + context_length).reshape(-1, 1),
            1,
            random_state=item_id + np.random.randint(100000),
        ).squeeze()

        support_set = gpr.sample_y(
            np.arange(context_length).reshape(-1, 1),
            support_set_size,
            random_state=item_id + np.random.randint(100000),
        ).T

        q = {
            "target": query.tolist(),
            "item_id": item_id,
            "start": "2012-01-01 00:00:00",
        }
        s = [
            {
                "target": s.tolist(),
                "start": "2012-01-01 00:00:00",
                "item_id": item_id,
            }
            for s in support_set
        ]
        return q, s



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# pylint: disable=missing-function-docstring
import logging
import pickle
import tempfile
from pathlib import Path
from typing import Any, Dict, Optional
import numpy as np
import torch
from sacred import Experiment
from tsbench.analysis import ModelRecommenderAnalyzer
from tsbench.constants import DEFAULT_DATA_PATH, DEFAULT_EVALUATIONS_PATH
from tsbench.evaluations.tracking import ModelTracker
from tsbench.recommender import create_recommender
from tsbench.surrogate import create_surrogate

ex = Experiment()


@ex.config
def experiment_config():
    # pylint: disable=unused-variable
    experiment = "test"  # type: ignore
    data_path = str(DEFAULT_DATA_PATH)  # type: ignore
    evaluations_path = str(DEFAULT_EVALUATIONS_PATH)  # type: ignore

    recommender = "pareto"  # type: ignore
    num_recommendations = 20  # type: ignore
    objectives = "ncrps_mean,latency_mean"  # type: ignore
    focus_objective = None  # type: ignore
    enforce_single_objective = False  # type: ignore

    surrogate = {  # type: ignore
        "name": "mlp",
        "inputs": {
            "use_simple_dataset_features": False,
            "use_seasonal_naive_performance": False,
            "use_catch22_features": False,
        },
        "outputs": {
            "normalization": "quantile",
            "imputation": False,
        },
        "xgboost": {
            "objective": "regression",
        },
        "autogluon": {
            "time_limit": 10,
        },
        "mlp": {
            "objective": "ranking",
            "discount": "linear",
            "hidden_layer_sizes": [32, 32],
            "weight_decay": 0.01,
            "dropout": 0.0,
        },
    }


@ex.automain
def main(
    _seed: int,
    data_path: str,
    evaluations_path: str,
    recommender: str,
    num_recommendations: int,
    objectives: str,
    focus_objective: Optional[str],
    enforce_single_objective: bool,
    surrogate: Dict[str, Any],
):
    assert (
        recommender != "surrogate" or surrogate["name"] is not None
    ), "Name of surrogate must be provided if surrogate recommender is used."

    np.random.seed(_seed)
    torch.manual_seed(_seed)
    logging.getLogger("pytorch_lightning").setLevel(logging.WARNING)

    # First, get the tracker
    print("Fetching the data...")
    tracker = ModelTracker.from_directory(
        Path(evaluations_path), data_path=Path(data_path)
    )

    # Then, potentially initialize the surrogate
    recommender_args: Dict[str, Any] = {
        "objectives": objectives.split(","),
        "focus": focus_objective,
    }

    if recommender == "pareto":
        print("Initializing the surrogate...")
        surrogate_metrics = [
            m
            for m in objectives.split(",")
            if (
                not m.startswith("latency")
                and not m.startswith("num_model_parameters")
            )
            or not surrogate["outputs"]["imputation"]
        ]
        recommender_args["surrogate"] = create_surrogate(
            surrogate["name"],
            predict=surrogate_metrics,
            tracker=tracker,
            input_flags=surrogate["inputs"],
            output_normalization=surrogate["outputs"]["normalization"],
            impute_simulatable=surrogate["outputs"]["imputation"],
            **(
                surrogate[surrogate["name"]]
                if surrogate["name"] in surrogate
                else {}
            ),
        )
    elif recommender == "optimal":
        recommender_args["tracker"] = tracker
    elif recommender == "greedy":
        recommender_args["enforce_single_objective"] = enforce_single_objective

    # Then, we can create the recommender
    print("Initializing the recommender...")
    recommender_instance = create_recommender(recommender, **recommender_args)

    # And evaluate it
    print("Evaluating the recommender...")
    evaluator = ModelRecommenderAnalyzer(
        tracker,
        recommender_instance,
        num_recommendations=num_recommendations,
    )
    recommendations = evaluator.run()

    # Eventually, we store the results
    print("Storing the results...")
    with tempfile.TemporaryDirectory() as d:
        path = Path(d) / "recommendations.pickle"
        with path.open("wb+") as f:
            pickle.dump(recommendations, f)
        ex.add_artifact(path, content_type="application/octet-stream")
